from datetime import timedelta

from django.test import TestCase
from django.utils import timezone

from hidp.federated import forms
from hidp.test.factories import user_factories


class TestOIDCRegistrationForm(TestCase):
    def _create_form(self, data=None, first_name="Firstname", last_name="Lastname"):
        return forms.OIDCRegistrationForm(
            data=data,
            provider_key="test_provider",
            claims={
                "iss": "test_issuer",
                "sub": "test_subject",
                "email": "user@example.com",
            },
            user_info={
                "given_name": first_name,
                "family_name": last_name,
            },
        )

    def test_data_from_claims_and_user_info(self):
        """The form is populated from the OIDC claims and user info."""
        form = self._create_form()
        self.assertEqual(form.initial["email"], "user@example.com")
        self.assertEqual(form.initial["first_name"], "Firstname")
        self.assertEqual(form.initial["last_name"], "Lastname")

    def test_partial_data_from_claims_and_user_info(self):
        """The form is populated from the OIDC claims and partial user info."""
        form = self._create_form(first_name="")
        self.assertEqual(form.initial["first_name"], "")
        self.assertEqual(form.initial["last_name"], "Lastname")

    def test_name_fields_hidden_when_data_missing(self):
        """First and last name fields are hidden when data is missing."""
        form = self._create_form(first_name="", last_name="")
        self.assertNotIn("first_name", form.fields)
        self.assertNotIn("last_name", form.fields)

    def test_tos_required(self):
        """Terms of service must be agreed to."""
        form = self._create_form(data={})
        self.assertFalse(form.is_valid(), msg="Expected form to be invalid.")
        self.assertFormError(form, "agreed_to_tos", "This field is required.")

    def test_email_field_disabled(self):
        """Email field is disabled to prevent the user from changing it."""
        form = self._create_form(
            data={
                "email": "fake@example.com",
                "agreed_to_tos": "on",
                "first_name": "Firstname",
                "last_name": "Lastname",
            }
        )
        self.assertTrue(form.is_valid(), msg="Expected form to be valid.")
        self.assertEqual(form.instance.email, "user@example.com")

    def test_creates_user_and_connection(self):
        """The form creates a user and OpenIdConnection from claims and user info."""
        form = self._create_form(
            data={
                "agreed_to_tos": "on",
                "first_name": "Firstname",
                "last_name": "Lastname",
            }
        )
        self.assertTrue(form.is_valid(), msg="Expected form to be valid.")
        user = form.save()
        self.assertAlmostEqual(
            user.agreed_to_tos, timezone.now(), delta=timedelta(seconds=10)
        )
        connection = user.openid_connections.first()
        self.assertIsNotNone(connection, msg="Expected connection to be created.")
        self.assertEqual(connection.provider_key, "test_provider")
        self.assertEqual(connection.issuer_claim, "test_issuer")
        self.assertEqual(connection.subject_claim, "test_subject")


class TestOIDCAccountLinkForm(TestCase):
    @classmethod
    def setUpTestData(cls):
        cls.user = user_factories.VerifiedUserFactory()

    def test_requires_confirmation(self):
        form = forms.OIDCAccountLinkForm(
            data={},
            user=self.user,
            provider_key="test_provider",
            claims={
                "iss": "test_issuer",
                "sub": "test_subject",
            },
        )
        self.assertFormError(form, "allow_link", "This field is required.")

    def test_creates_connection(self):
        form = forms.OIDCAccountLinkForm(
            data={"allow_link": "on"},
            user=self.user,
            provider_key="test_provider",
            claims={
                "iss": "test_issuer",
                "sub": "test_subject",
            },
        )
        self.assertTrue(form.is_valid(), msg="Expected form to be valid.")
        form.save()
        connection = self.user.openid_connections.first()
        self.assertIsNotNone(connection, msg="Expected connection to be created.")



import logging

from datetime import timedelta
from urllib.parse import urlencode

from django_ratelimit.decorators import ratelimit

from django.conf import settings
from django.contrib.auth import get_user_model
from django.contrib.auth import views as auth_views
from django.contrib.auth.mixins import LoginRequiredMixin
from django.core.exceptions import PermissionDenied
from django.db import IntegrityError
from django.db.models.functions import MD5
from django.http import Http404, HttpResponseRedirect
from django.shortcuts import resolve_url
from django.urls import reverse, reverse_lazy
from django.utils import timezone
from django.utils.decorators import method_decorator
from django.utils.translation import gettext_lazy as _
from django.views import generic
from django.views.decorators.cache import never_cache

from hidp.utils import get_account_management_links, is_registration_enabled

from ..config import oidc_clients
from ..csp.decorators import hidp_csp_protection
from ..federated.views import OIDCContextMixin
from ..otp.decorators import otp_exempt
from ..rate_limit.decorators import rate_limit_default, rate_limit_strict
from . import auth as hidp_auth
from . import email_verification, forms, mailers, tokens
from .email_change import Recipient
from .models import EmailChangeRequest

logger = logging.getLogger(__name__)
UserModel = get_user_model()


@method_decorator(hidp_csp_protection, name="dispatch")
@method_decorator(ratelimit(key="ip", rate="2/s", method="POST"), name="post")
@method_decorator(ratelimit(key="ip", rate="5/m", method="POST"), name="post")
@method_decorator(ratelimit(key="ip", rate="30/15m", method="POST"), name="post")
class RegistrationView(auth_views.RedirectURLMixin, OIDCContextMixin, generic.FormView):
    """
    Display the registration form and handle the registration action.

    If the form is submitted with valid data, a new user account will be created
    and the user will be redirected to a page informing them that they must verify
    their email address.

    Otherwise, the form will be displayed with an error message explaining the
    reason for the failure and the user can try again.
    """

    form_class = forms.UserCreationForm
    template_name = "hidp/accounts/register.html"
    next_page = "/"
    verification_mailer = mailers.EmailVerificationMailer
    account_exists_mailer = mailers.AccountExistsMailer

    def _build_provider_url_list(
        self,
        providers,
        url_name="hidp_oidc_client:authenticate",
        label=None,
    ):
        return super()._build_provider_url_list(
            providers,
            url_name,
            label=label or _("Sign up using {provider}"),
        )

    def get_context_data(self, **kwargs):
        login_url = resolve_url(settings.LOGIN_URL) + (
            f"?{urlencode({'next': redirect_url})}"
            if (redirect_url := self.get_redirect_url())
            else ""
        )
        context = {
            "user": self.request.user,
            "login_url": login_url,
            "next": self.get_success_url(),
            "logout_url": reverse("hidp_accounts:logout"),
            "logout_next_url": self.request.get_full_path(),
            "can_register": not self.request.user.is_authenticated,
        }
        return super().get_context_data() | context | kwargs

    def dispatch(self, request, *args, **kwargs):
        if not is_registration_enabled():
            raise Http404("Registration is disabled.")
        return super().dispatch(request, *args, **kwargs)

    def post(self, request, *args, **kwargs):
        if request.user.is_authenticated:
            raise PermissionDenied("Logged-in users cannot register a new account.")
        return super().post(request, *args, **kwargs)

    def send_email(self, user):
        """Send the appropriate email to the user."""
        base_url = self.request.build_absolute_uri("/")

        try:
            if not user.email_verified:
                self.verification_mailer(
                    user,
                    base_url=base_url,
                    post_verification_redirect=self.get_redirect_url(),
                ).send()
            elif user.is_active:
                # Email the user to inform them that they have an account.
                self.account_exists_mailer(
                    user,
                    base_url=base_url,
                ).send()
        except Exception:
            # Do not leak the existence of the user. Log the error and
            # continue as if the email was sent successfully.
            logger.exception("Failed to send verification email.")

    def form_valid(self, form):
        """Save the new user and redirect to the email verification required page."""
        try:
            user = form.save()
        except IntegrityError:
            # The user exists! Find the user by the email address (case-insensitive).
            user = UserModel.objects.get(email__iexact=form.cleaned_data["email"])

        self.send_email(user)

        # Always redirect to the email verification required page.
        # This is a security measure to prevent user enumeration.
        return HttpResponseRedirect(
            email_verification.get_email_verification_required_url(
                user, next_url=self.get_redirect_url()
            )
        )


@method_decorator(hidp_csp_protection, name="dispatch")
class TermsOfServiceView(generic.TemplateView):
    """Display the terms of service."""

    template_name = "hidp/accounts/tos.html"


class BaseTokenMixin:
    """Mixin to handle tokens in URLs."""

    token_generator = NotImplemented
    token_session_key = NotImplemented
    token_placeholder = "token"  # noqa: S105 (not a password)
    template_name_invalid_link = NotImplemented

    def _remove_token_from_url(self, token):
        """
        Move the token from the URL to the session and redirect to the placeholder URL.

        If the url already is the placeholder URL, do nothing.
        """
        if token == self.token_placeholder:
            # Token is already the placeholder value, so do nothing.
            return None
        # Store the token in the session and redirect to the
        # URL with a placeholder value.
        self.request.session[self.token_session_key] = token
        redirect_url = self.request.get_full_path().replace(
            token, self.token_placeholder
        )
        return HttpResponseRedirect(redirect_url, status=307)

    def get_template_names(self):
        """Return the template names to use for the view."""
        if self.validlink:
            return super().get_template_names()
        else:
            return [self.template_name_invalid_link]

    def validate_token_data(self, token_data):  # noqa: PLR6301 (no-self-use)
        """
        Validate the token data.

        Override this method to add additional validation checks.

        Returns:
            bool:
                True if the token data is valid, otherwise False.
        """
        return token_data is not None

    def dispatch(self, request, *, token):
        """
        Handle tokens in URLs.

        Makes sure the token is removed from the URL and stored in
        the session.

        Sets the `validlink` attribute to whether the token is valid.

        If `validlink` is False, the `get` method will be called
        irrespective of the HTTP method used.
        """
        response = self._remove_token_from_url(token)
        if response:
            return response

        token = self.request.session.get(self.token_session_key)
        token_data = self.token_generator.check_token(token) if token else None
        self.validlink = self.validate_token_data(token_data)

        if not self.validlink:
            # Invalid token, render the invalid link template.
            return self.render_to_response({})

        return super().dispatch(request, token=token)


class EmailTokenMixin(BaseTokenMixin):
    """Mixin to handle email verification tokens in URLs."""

    token_placeholder = "email"  # noqa: S105 (not a password)

    def _get_user_queryset(self):  # noqa: PLR6301 (no-self-use)
        """
        Base queryset for finding the user by the token.

        Override this method to customize the queryset, i.e. to
        add additional filters or annotations.
        """
        return UserModel.objects.annotate(email_hash=MD5("email"))

    def _get_user_from_token(self, email_hash):
        """
        Find the user associated with the token in the session.

        Returns:
            UserModel | None:
                The user if the token is valid, otherwise None.
        """
        # Find the user by the hash of their email address
        return self._get_user_queryset().filter(email_hash=email_hash).first()

    def validate_token_data(self, token_data):
        """
        Validate the token data.

        Sets the `user` attribute to the user found by the token,
        or `None` if the token is invalid.

        Returns:
            bool:
                True if the token data is valid, otherwise False.
        """
        if super().validate_token_data(token_data):
            self.user = self._get_user_from_token(token_data)
            return self.user is not None
        else:
            self.user = None
            return False


class EmailChangeTokenMixin(BaseTokenMixin):
    """Mixin to handle email change tokens in URLs."""

    token_placeholder = "email-change"  # noqa: S105 (not a password)

    def _get_email_change_request_from_token_object(self, token_object):
        """
        Find the email change request associated with the token in the session.

        Exclude the request if it has already been confirmed for this email address.

        Returns:
            EmailChangeRequest | None:
                The email change request if the token is valid, otherwise None.
        """
        email_change_request = (
            EmailChangeRequest.objects.filter(id=token_object["uuid"])
            .exclude(**{f"confirmed_by_{token_object['recipient']}": True})
            .first()
        )

        if (
            email_change_request is None
            or email_change_request.user != self.request.user
        ):
            return None
        return email_change_request

    def validate_token_data(self, token_data):
        """
        Validate the token data.

        Sets the `email_change_request` attribute to the email change request
        found by the token, or `None` if the token is invalid.

        Sets the `recipient` attribute to the recipient of the token,
        or `None` if the token is invalid.

        Returns:
            bool:
                True if the token data is valid, otherwise False.
        """
        if super().validate_token_data(token_data):
            self.email_change_request = (
                self._get_email_change_request_from_token_object(token_data)
            )
            self.recipient = token_data["recipient"]
            return self.email_change_request is not None
        else:
            self.email_change_request = None
            self.recipient = None
            return False


@method_decorator(hidp_csp_protection, name="dispatch")
@method_decorator(rate_limit_default, name="dispatch")
@method_decorator(never_cache, name="dispatch")
class EmailVerificationRequiredView(
    auth_views.RedirectURLMixin,
    EmailTokenMixin,
    generic.TemplateView,
):
    """
    Display a notice that the user must verify their email address.

    Can be used to resend the email verification email by sending a POST request.
    """

    template_name = "hidp/accounts/verification/email_verification_required.html"
    template_name_invalid_link = (
        "hidp/accounts/verification/email_verification_required_invalid_link.html"
    )
    token_generator = tokens.email_verification_request_token_generator
    verification_mailer = mailers.EmailVerificationMailer
    token_session_key = "_email_verification_request_token"  # noqa: S105 (not a password)

    def send_email(self):
        """Send the email verification email."""
        self.verification_mailer(
            self.user,
            base_url=self.request.build_absolute_uri("/"),
            post_verification_redirect=self.get_redirect_url(),
        ).send()

    def post(self, *args, **kwargs):
        if self.validlink:
            # Send the email verification email.
            self.send_email()
            # Redirect to the email verification required page, with a new token.
            return HttpResponseRedirect(
                email_verification.get_email_verification_required_url(
                    self.user, next_url=self.get_redirect_url()
                )
            )
        # Invalid token, do nothing and redirect to the same page.
        return HttpResponseRedirect(self.request.get_full_path())


@method_decorator(hidp_csp_protection, name="dispatch")
@method_decorator(rate_limit_default, name="dispatch")
@method_decorator(never_cache, name="dispatch")
class EmailVerificationView(
    auth_views.RedirectURLMixin,
    EmailTokenMixin,
    generic.UpdateView,
):
    """
    Landing page for email verification links.

    Contains a form that must be submitted to complete the verification process.
    """

    form_class = forms.EmailVerificationForm
    template_name = "hidp/accounts/verification/verify_email.html"
    template_name_invalid_link = (
        "hidp/accounts/verification/verify_email_invalid_link.html"
    )
    token_generator = tokens.email_verification_token_generator
    success_url = reverse_lazy("hidp_accounts:email_verification_complete")
    token_session_key = "_email_verification_request_token"  # noqa: S105 (not a password)

    def _get_user_queryset(self):
        return super()._get_user_queryset().email_unverified().filter(is_active=True)

    def get_object(self):
        return self.user  # The user from the token

    def form_valid(self, form):
        form.save()
        return HttpResponseRedirect(
            str(self.success_url)
            + (
                f"?{urlencode({'next': redirect_url})}"
                if (redirect_url := self.get_redirect_url())
                else ""
            )
        )


@method_decorator(hidp_csp_protection, name="dispatch")
class EmailVerificationCompleteView(auth_views.RedirectURLMixin, generic.TemplateView):
    """Display a message that the email address has been verified."""

    template_name = "hidp/accounts/verification/email_verification_complete.html"

    def get_context_data(self, **kwargs):
        login_url = resolve_url(settings.LOGIN_URL) + (
            f"?{urlencode({'next': redirect_url})}"
            if (redirect_url := self.get_redirect_url())
            else ""
        )
        context = {
            "login_url": login_url,
        }
        return super().get_context_data() | context | kwargs


@method_decorator(hidp_csp_protection, name="dispatch")
@method_decorator(
    ratelimit(key="post:username", rate="10/m", method="POST", block=False), name="post"
)
@method_decorator(rate_limit_strict, name="dispatch")
class LoginView(OIDCContextMixin, auth_views.LoginView):
    """
    Display the login form and handle the login action.

    If the form is submitted with valid credentials, the user will be logged in
    and redirected to the location returned by get_success_url().

    Otherwise, the form will be displayed with an error message explaining the
    reason for the failure and the user can try again.
    """

    # The form class to use for authentication
    form_class = forms.AuthenticationForm

    # The form class to use when the user is rate limited
    rate_limited_form_class = forms.RateLimitedAuthenticationForm

    # The template to use for displaying the login form
    template_name = "hidp/accounts/login.html"

    # If the user is already authenticated, redirect to the success URL
    # instead of displaying the login form.
    redirect_authenticated_user = False

    # Mailer class to use when a user's email address is not verified
    verification_mailer = mailers.EmailVerificationMailer

    def get_context_data(self, **kwargs):
        """
        Additional context data for the login template.

        By default, the context data includes:

        * `view`: The current view instance
        * `form`: The login form
        * `self.redirect_field_name` (i.e. `next`):
          The URL to redirect to after login (if present in the request)
        * `site`:
          The current site instance
          (`RequestSite` if `django.contrib.sites` is not installed)
        * `site_name`:
          The name of the current site (host name if `RequestSite` is used)
        * Any additional data present is `self.extra_context`
        """
        register_url = None

        if is_registration_enabled():
            register_url = reverse("hidp_accounts:register") + (
                f"?{urlencode({'next': redirect_url})}"
                if (redirect_url := self.get_redirect_url())
                else ""
            )
        context = {
            "password_reset_url": reverse("hidp_accounts:password_reset_request"),
            "register_url": register_url,
            "is_rate_limited": self.request.limited,
        }
        return super().get_context_data() | context | kwargs

    def get_success_url(self):
        """
        Return the URL to redirect to after a successful login.

        Returns one of the following:

        1. The URL specified by the `self.redirect_field_name`
          (i.e. `next`) parameter, if it is present in the request and
          the value is valid and safe.
        2. The URL specified by `self.next_page` if it is set.
        3. `settings.LOGIN_REDIRECT_URL` if it is set.
        """
        return super().get_success_url()

    def get_form_class(self):
        """
        Determine the form class to use for the view.

        If the request is rate limited, return a form that requires the user to prove
        they are not a bot.
        Otherwise, return the normal authentication form.
        """
        if self.request.limited:
            return self.rate_limited_form_class
        return super().get_form_class()

    def send_email(self, user):
        """Send the email verification email."""
        self.verification_mailer(
            user,
            base_url=self.request.build_absolute_uri("/"),
            post_verification_redirect=self.get_redirect_url(),
        ).send()

    def form_valid(self, form):
        """
        User has provided valid credentials and is allowed to log in.

        Persist the user and backend in the session and redirect to the
        success URL.

        If the user's email address has not been verified, redirect them
        to the email verification required flow.
        """
        user = form.get_user()
        if user.email_verified:
            # Only log in the user if their email address has been verified.
            hidp_auth.login(self.request, user)
            return HttpResponseRedirect(self.get_success_url())

        # If the user's email address is not yet verified:
        # Send the email verification email.
        self.send_email(user)

        # Then redirect them to the email verification required page.
        return HttpResponseRedirect(
            email_verification.get_email_verification_required_url(
                user, next_url=self.get_redirect_url()
            )
        )


@method_decorator(otp_exempt, name="dispatch")
@method_decorator(rate_limit_default, name="dispatch")
class LogoutView(auth_views.LogoutView):
    """
    Logs out the user, regardless of whether a user is logged in.

    A POST request (including a CSRF token) is required to log out.
    This prevents a malicious site from logging out a user without their consent,
    for example by linking to the logout URL.

    After logging out, the user is redirected to the URL returned by get_redirect_url().
    """

    # Django 5.0 will no longer allow GET (and HEAD) requests to the logout view.
    # Disallow these methods now for forward compatibility.
    http_method_names = [
        method
        for method in auth_views.LogoutView.http_method_names
        if method not in {"get", "head"}
    ]

    def get_redirect_url(self):
        """
        Return the URL to redirect to after a successful logout.

        Returns one of the following:

        1. The URL specified by the `self.redirect_field_name`
          (i.e. `next`) parameter, if it is present in the request and
          the value is valid and safe.
        2. The URL specified by `self.next_page` if it is set.
        3. `settings.LOGOUT_REDIRECT_URL` if it is set.
        """
        return super().get_redirect_url()

    def post(self, request, *args, **kwargs):
        """Log out the user and redirect to the success URL."""
        # This **replaces** the base implementation in order to use the
        # HIdP logout wrapper, for good measure.
        hidp_auth.logout(request)
        redirect_to = self.get_success_url()
        if redirect_to != request.get_full_path():
            # Redirect to target page once the session has been cleared.
            return HttpResponseRedirect(redirect_to)
        return super().get(request, *args, **kwargs)


@method_decorator(hidp_csp_protection, name="dispatch")
@method_decorator(rate_limit_strict, name="dispatch")
class PasswordResetRequestView(generic.FormView):
    """
    Display and handle the password reset request form.

    Sends the password reset email and redirects to the password reset
    sent view if the form is submitted with valid data.
    """

    form_class = forms.PasswordResetRequestForm
    template_name = "hidp/accounts/recovery/password_reset_request.html"
    success_url = reverse_lazy("hidp_accounts:password_reset_email_sent")
    password_reset_request_mailer = mailers.PasswordResetRequestMailer
    set_password_mailer = mailers.SetPasswordMailer

    def get_context_data(self, **kwargs):
        context = {
            "cancel_url": resolve_url(settings.LOGIN_URL),
        }
        return super().get_context_data() | context | kwargs

    def send_email(self, user):
        """Send the appropriate email to the user."""
        if user.has_usable_password():
            mailer_class = self.password_reset_request_mailer
        else:
            mailer_class = self.set_password_mailer
        try:
            mailer_class(
                user=user,
                base_url=self.request.build_absolute_uri("/"),
            ).send()
        except Exception:
            # Do not leak the existence of the user. Log the error and
            # continue as if the email was sent successfully.
            logger.exception("Failed to send password (re)set email.")

    def form_valid(self, form):
        if user := form.get_user():
            self.send_email(user)
        return super().form_valid(form)


@method_decorator(hidp_csp_protection, name="dispatch")
class PasswordResetEmailSentView(generic.TemplateView):
    """Display a message that the password reset email has been sent."""

    template_name = "hidp/accounts/recovery/password_reset_email_sent.html"


@method_decorator(hidp_csp_protection, name="dispatch")
@method_decorator(rate_limit_default, name="dispatch")
class PasswordResetView(auth_views.PasswordResetConfirmView):
    """Display the password reset form and handle the password reset action."""

    form_class = forms.PasswordResetForm
    template_name = "hidp/accounts/recovery/password_reset.html"
    template_name_invalid_link = (
        "hidp/accounts/recovery/password_reset_invalid_link.html"
    )
    success_url = reverse_lazy("hidp_accounts:password_reset_complete")
    password_changed_mailer = mailers.PasswordChangedMailer

    def get_template_names(self):
        """Return the template names to use for the view."""
        if self.validlink:
            return super().get_template_names()
        else:
            return [self.template_name_invalid_link]

    def send_email(self):
        """Send the password changed email."""
        self.password_changed_mailer(
            self.user,
            base_url=self.request.build_absolute_uri("/"),
        ).send()

    def form_valid(self, form):
        response = super().form_valid(form)
        self.send_email()
        return response


@method_decorator(hidp_csp_protection, name="dispatch")
class PasswordResetCompleteView(auth_views.TemplateView):
    """Display a message that the password reset has been completed."""

    template_name = "hidp/accounts/recovery/password_reset_complete.html"

    def get_context_data(self, **kwargs):
        context = {
            "login_url": resolve_url(settings.LOGIN_URL),
        }
        return super().get_context_data() | context | kwargs


@method_decorator(hidp_csp_protection, name="dispatch")
@method_decorator(rate_limit_default, name="dispatch")
class PasswordChangeView(LoginRequiredMixin, auth_views.PasswordChangeView):
    """Display the password change form and handle the password change action."""

    form_class = forms.PasswordChangeForm
    template_name = "hidp/accounts/management/password_change.html"
    success_url = reverse_lazy("hidp_account_management:change_password_done")
    password_changed_mailer = mailers.PasswordChangedMailer

    def dispatch(self, request, *args, **kwargs):
        if request.user.is_authenticated and not request.user.has_usable_password():
            return HttpResponseRedirect(reverse("hidp_account_management:set_password"))
        return super().dispatch(request, *args, **kwargs)

    def get_context_data(self, **kwargs):
        context = {
            "cancel_url": reverse("hidp_account_management:manage_account"),
        }
        return super().get_context_data() | context | kwargs

    def send_email(self):
        """Send the password changed email."""
        self.password_changed_mailer(
            self.request.user,
            base_url=self.request.build_absolute_uri("/"),
        ).send()

    def form_valid(self, form):
        response = super().form_valid(form)
        self.send_email()
        return response


@method_decorator(hidp_csp_protection, name="dispatch")
class PasswordChangeDoneView(auth_views.TemplateView):
    """Display a message that the password change has been completed."""

    template_name = "hidp/accounts/management/password_change_done.html"

    def get_context_data(self, **kwargs):
        context = {
            "back_url": reverse("hidp_account_management:manage_account"),
        }
        return super().get_context_data() | context | kwargs


@method_decorator(hidp_csp_protection, name="dispatch")
@method_decorator(rate_limit_default, name="dispatch")
class SetPasswordView(
    LoginRequiredMixin, OIDCContextMixin, auth_views.PasswordChangeView
):
    """Allow users without a password to set one."""

    form_class = forms.SetPasswordForm
    template_name = "hidp/accounts/management/set_password.html"
    success_url = reverse_lazy("hidp_account_management:set_password_done")
    login_delta = timedelta(minutes=5)
    password_changed_mailer = mailers.PasswordChangedMailer

    def dispatch(self, request, *args, **kwargs):
        if not request.user.is_authenticated:
            return self.handle_no_permission()

        if request.user.has_usable_password():
            return HttpResponseRedirect(
                reverse_lazy("hidp_account_management:change_password")
            )

        last_login = request.user.last_login
        # If the user has not logged in recently, they must re-authenticate
        # to prove their identity before setting a password.
        self.must_reauthenticate = last_login is None or last_login < (
            timezone.now() - self.login_delta
        )

        return super().dispatch(request, *args, **kwargs)

    def _get_linked_oidc_providers(self):
        """User's linked OIDC providers in client registration order."""
        linked_provider_keys = self.request.user.openid_connections.values_list(
            "provider_key", flat=True
        )
        for provider in oidc_clients.get_registered_oidc_clients():
            if provider.provider_key in linked_provider_keys:
                yield provider

    def get_context_data(self, **kwargs):
        context = {
            "cancel_url": reverse("hidp_account_management:manage_account"),
            "must_reauthenticate": self.must_reauthenticate,
            "oidc_linked_providers": self._build_provider_url_list(
                self._get_linked_oidc_providers() if self.must_reauthenticate else (),
                url_name="hidp_oidc_client:reauthenticate",
                label=_("Authenticate with {provider}"),
            ),
            "auth_next_url": self.request.get_full_path(),
        }
        return super().get_context_data() | context | kwargs

    def post(self, request, *args, **kwargs):
        if self.must_reauthenticate:
            # The user was able to POST the form, but has not logged in recently.
            # Redirect to this view using a GET so they are shown the message
            # that they must re-authenticate.
            return HttpResponseRedirect(reverse("hidp_account_management:set_password"))
        return super().post(request, *args, **kwargs)

    def send_email(self):
        """Send the password changed email."""
        self.password_changed_mailer(
            self.request.user,
            base_url=self.request.build_absolute_uri("/"),
        ).send()

    def form_valid(self, form):
        form.save()
        self.send_email()
        return super().form_valid(form)


@method_decorator(hidp_csp_protection, name="dispatch")
class SetPasswordDoneView(auth_views.TemplateView):
    """Display a message that the password has been set."""

    template_name = "hidp/accounts/management/set_password_done.html"

    def get_context_data(self, **kwargs):
        context = {
            "back_url": reverse("hidp_account_management:manage_account"),
        }
        return super().get_context_data() | context | kwargs


@method_decorator(hidp_csp_protection, name="dispatch")
class ManageAccountView(LoginRequiredMixin, OIDCContextMixin, generic.TemplateView):
    """Display the manage account page."""

    template_name = "hidp/accounts/management/manage_account.html"

    def get_context_data(self, **kwargs):
        context = {
            "user": self.request.user,
            "logout_url": reverse("hidp_accounts:logout"),
            "account_management_links": get_account_management_links(self.request.user),
        }
        return super().get_context_data() | context | kwargs


@method_decorator(hidp_csp_protection, name="dispatch")
@method_decorator(rate_limit_default, name="dispatch")
class EditAccountView(LoginRequiredMixin, generic.FormView):
    """Display the edit user form and handle the edit user action."""

    template_name = "hidp/accounts/management/edit_account.html"
    form_class = forms.EditUserForm
    success_url = reverse_lazy("hidp_account_management:edit_account_done")

    def get_context_data(self, **kwargs):
        context = {
            "cancel_url": reverse("hidp_account_management:manage_account"),
        }
        return super().get_context_data() | context | kwargs

    def get_form_kwargs(self):
        return super().get_form_kwargs() | {
            "instance": self.request.user,
        }

    def form_valid(self, form):
        form.save()
        return super().form_valid(form)


@method_decorator(hidp_csp_protection, name="dispatch")
class EditAccountDoneView(generic.TemplateView):
    """Display a message that the account has been updated."""

    template_name = "hidp/accounts/management/edit_account_done.html"

    def get_context_data(self, **kwargs):
        context = {
            "back_url": reverse("hidp_account_management:manage_account"),
        }
        return super().get_context_data() | context | kwargs


@method_decorator(hidp_csp_protection, name="dispatch")
@method_decorator(rate_limit_default, name="dispatch")
class EmailChangeRequestView(LoginRequiredMixin, generic.CreateView):
    """
    Display and handle the email change request form.

    If the form is submitted with valid data, emails are sent to the user's
    current and proposed email address to confirm the change request.
    The user is then redirected to the success page.
    """

    form_class = forms.EmailChangeRequestForm
    template_name = "hidp/accounts/management/email_change_request.html"
    success_url = reverse_lazy("hidp_account_management:email_change_request_sent")
    email_change_request_mailer = mailers.EmailChangeRequestMailer
    proposed_email_exists_mailer = mailers.ProposedEmailExistsMailer

    def get_form_kwargs(self):
        return {
            **super().get_form_kwargs(),
            "user": self.request.user,
        }

    def get_context_data(self, **kwargs):
        context = {
            "can_change_email": self.request.user.has_usable_password(),
            "set_password_url": reverse("hidp_account_management:set_password"),
            "cancel_url": reverse("hidp_account_management:manage_account"),
        }
        return super().get_context_data() | context | kwargs

    def send_email(self, email_change_request):
        """Send the email change confirmation emails."""
        mailer_kwargs = {
            "user": self.request.user,
            "email_change_request": email_change_request,
            "base_url": self.request.build_absolute_uri("/"),
        }
        self.email_change_request_mailer(
            **mailer_kwargs,
            recipient=Recipient.CURRENT_EMAIL,
        ).send()

        proposed_email_mailer_class = self.email_change_request_mailer
        existing_user = UserModel.objects.filter(
            email__iexact=email_change_request.proposed_email
        ).first()

        if existing_user and not existing_user.is_active:
            # Do nothing if the user exists but is not active.
            return

        if existing_user:
            # Send an email to the proposed email address to inform them that
            # an account with this email address already exists.
            proposed_email_mailer_class = self.proposed_email_exists_mailer

        proposed_email_mailer_class(
            **mailer_kwargs,
            recipient=Recipient.PROPOSED_EMAIL,
        ).send()

    def form_valid(self, form):
        email_change_request = form.save()
        self.send_email(email_change_request)
        return HttpResponseRedirect(self.success_url)


@method_decorator(hidp_csp_protection, name="dispatch")
class EmailChangeRequestSentView(generic.TemplateView):
    """Display a message that the email change request confirmation emails were sent."""

    template_name = "hidp/accounts/management/email_change_request_sent.html"

    def get_context_data(self, **kwargs):
        context = {
            "back_url": reverse("hidp_account_management:manage_account"),
        }
        return super().get_context_data() | context | kwargs


@method_decorator(hidp_csp_protection, name="dispatch")
@method_decorator(rate_limit_default, name="dispatch")
class EmailChangeConfirmView(
    LoginRequiredMixin,
    EmailChangeTokenMixin,
    generic.UpdateView,
):
    """
    Landing page for email change confirmation links.

    Contains a form that must be submitted to complete the email change process.
    """

    form_class = forms.EmailChangeConfirmForm
    template_name = "hidp/accounts/management/email_change_confirm.html"
    template_name_invalid_link = (
        "hidp/accounts/management/email_change_confirm_invalid_link.html"
    )
    success_url = reverse_lazy("hidp_account_management:email_change_complete")
    token_generator = tokens.email_change_token_generator
    token_session_key = "_email_change_request_token"  # noqa: S105 (not a password)
    email_changed_mailer = mailers.EmailChangedMailer

    def get_context_data(self, **kwargs):
        context = {
            "cancel_url": reverse("hidp_account_management:manage_account"),
            "recipient": self.recipient,
            "current_email": self.email_change_request.current_email,
            "proposed_email": self.email_change_request.proposed_email,
        }
        return super().get_context_data() | context | kwargs

    def get_form_kwargs(self):
        return super().get_form_kwargs() | {
            "recipient": self.recipient,
        }

    def get_object(self):
        return self.email_change_request

    def form_valid(self, form):
        try:
            form.save()
        except IntegrityError:
            form.add_error(
                None,
                _(
                    "Sorry, changing your email address is not possible because an"
                    " account with this email address already exists."
                ),
            )
            return self.form_invalid(form)
        if self.email_change_request.is_complete():
            self.send_email()
        return HttpResponseRedirect(self.success_url)

    def send_email(self):
        """Send the email changed email."""
        self.email_changed_mailer(
            self.request.user,
            email_change_request=self.email_change_request,
            base_url=self.request.build_absolute_uri("/"),
        ).send()


@method_decorator(hidp_csp_protection, name="dispatch")
class EmailChangeCompleteView(auth_views.TemplateView):
    """Display a message that the email change has been completed."""

    template_name = "hidp/accounts/management/email_change_complete.html"

    def get_context_data(self, **kwargs):
        email_change_request = EmailChangeRequest.objects.filter(
            user=self.request.user
        ).first()
        context = {
            "back_url": reverse("hidp_account_management:manage_account"),
        }
        if email_change_request is None or email_change_request.is_complete():
            context |= {
                "current_email_confirmation_required": False,
                "proposed_email_confirmation_required": False,
                "email_change_request_completed": True,
            }
        else:
            context |= {
                "current_email_confirmation_required": (
                    not email_change_request.confirmed_by_current_email
                ),
                "proposed_email_confirmation_required": (
                    not email_change_request.confirmed_by_proposed_email
                ),
                "email_change_request_completed": False,
            }
        return super().get_context_data() | context | kwargs


@method_decorator(hidp_csp_protection, name="dispatch")
@method_decorator(rate_limit_default, name="dispatch")
class EmailChangeCancelView(LoginRequiredMixin, generic.DeleteView):
    """Cancel an email change request."""

    form_class = forms.EmailChangeCancelForm
    template_name = "hidp/accounts/management/email_change_cancel.html"
    template_name_invalid_link = (
        "hidp/accounts/management/email_change_cancel_invalid_link.html"
    )
    success_url = reverse_lazy("hidp_account_management:email_change_cancel_done")
    # This view does not use a token. The token generator is only used
    # to limit the change request lookup to those that have not expired.
    token_generator = tokens.email_change_token_generator

    def dispatch(self, request, *args, **kwargs):
        if not request.user.is_authenticated:
            return self.handle_no_permission()
        self.object = self.get_object()
        if not self.object:
            # No object found, show invalid or expired link message.
            return self.render_to_response({})
        return super().dispatch(request, *args, **kwargs)

    def get_template_names(self):
        if self.object:
            return super().get_template_names()
        else:
            # No object found, show invalid or expired link message.
            return [self.template_name_invalid_link]

    def get_context_data(self, **kwargs):
        context = {
            "current_email": self.object.current_email,
            "proposed_email": self.object.proposed_email,
            "cancel_url": reverse("hidp_account_management:manage_account"),
        }
        return super().get_context_data() | context | kwargs

    def get_object(self, queryset=None):
        """
        Get the email change request to cancel.

        But only if there is a request for the current user that has not been confirmed
        by both the current and proposed email addresses, and has not expired.
        """
        if hasattr(self, "object"):
            # To avoid duplicate queries in the get and post handlers, return
            # the object that was already retrieved in the dispatch method.
            return self.object
        return (
            EmailChangeRequest.objects.filter(
                user=self.request.user,
                created_at__gte=(
                    timezone.now()
                    - timedelta(seconds=self.token_generator.token_timeout)
                ),
            )
            .exclude(
                confirmed_by_current_email=True,
                confirmed_by_proposed_email=True,
            )
            .first()
        )


@method_decorator(hidp_csp_protection, name="dispatch")
class EmailChangeCancelDoneView(auth_views.TemplateView):
    """Display a message that the email change request has been cancelled."""

    template_name = "hidp/accounts/management/email_change_cancel_done.html"

    def get_context_data(self, **kwargs):
        context = {
            "back_url": reverse("hidp_account_management:manage_account"),
        }
        return super().get_context_data() | context | kwargs



import io

from http import HTTPStatus
from unittest import mock

from django.core import mail
from django.core.management import call_command
from django.test import TestCase
from django.urls import reverse
from django.utils import timezone

from hidp.accounts import tokens
from hidp.accounts.email_change import remove_complete_and_stale_email_change_requests
from hidp.accounts.forms import EmailChangeRequestForm
from hidp.accounts.models import EmailChangeRequest
from hidp.test.factories import user_factories


class TestEmailChangeRequest(TestCase):
    @classmethod
    def setUpTestData(cls):
        cls.user = user_factories.UserFactory()
        cls.url = reverse("hidp_account_management:email_change_request")

    def setUp(self):
        self.client.force_login(self.user)

    def test_get_unauthenticated(self):
        self.client.logout()
        response = self.client.get(self.url)

        self.assertEqual(response.status_code, 302)

    def test_get(self):
        response = self.client.get(self.url)

        self.assertEqual(response.status_code, 200)
        self.assertTemplateUsed(
            response, "hidp/accounts/management/email_change_request.html"
        )

    def test_get_user_without_password_requests_email_change(self):
        self.user.set_unusable_password()
        self.user.save()

        self.client.force_login(self.user)

        response = self.client.get(self.url)
        self.assertInHTML(
            "Your account does not currently have a password set.",
            response.content.decode(),
        )

    def test_post_user_without_password_requests_email_change(self):
        self.user.set_unusable_password()
        self.user.save()

        self.client.force_login(self.user)
        response = self.client.post(
            self.url,
            {
                "password": "P@ssw0rd!",
                "proposed_email": "newemail@example.com",
            },
        )
        self.assertFalse(response.context["form"].is_valid())
        self.assertIn("password", response.context["form"].errors)

    def test_user_requests_email_change(self):
        with (
            self.assertTemplateUsed("hidp/accounts/management/email/email_change_subject.txt"),
            self.assertTemplateUsed("hidp/accounts/management/email/email_change_body.txt"),
            self.assertTemplateUsed("hidp/accounts/management/email/email_change_body.html"),
        ):  # fmt: skip
            response = self.client.post(
                self.url,
                {
                    "password": "P@ssw0rd!",
                    "proposed_email": "newemail@example.com",
                },
                follow=True,
            )

        self.assertRedirects(
            response, reverse("hidp_account_management:email_change_request_sent")
        )
        self.assertTemplateUsed(
            response, "hidp/accounts/management/email_change_request_sent.html"
        )

        # EmailChangeRequest should be created
        self.assertTrue(
            EmailChangeRequest.objects.filter(
                user=self.user,
                proposed_email="newemail@example.com",
            ).exists()
        )

        # Email should be sent to current email
        self.assertEqual(len(mail.outbox), 2)

        message = mail.outbox[0]
        self.assertEqual(
            message.subject,
            "Confirm your email change request",
        )
        self.assertEqual(message.to, [self.user.email])
        self.assertRegex(
            message.body,
            # Matches the email change confirmation URL:
            # http://testserver/manage/change-email-confirm/eyJ1dWlkIjoiMDE5MjZiNGYtODQ0Zi03MjRmLWE2YjQtMWQxYWEyYTU5OTgwIiwicmVjaXBpZW50IjoiY3VycmVudF9lbWFpbCJ9:1sy5S2:R7m51osUdabcMuOGXZRq7MabESIqKGl_mX2jO-TAcj8/
            r"http://testserver/manage/change-email-confirm/[0-9A-Za-z]+:[0-9a-zA-Z]+:[0-9A-Za-z_-]+/",
        )
        self.assertIn(
            "http://testserver/manage/change-email-cancel/",
            message.body,
        )

        # Email should be sent to proposed email
        message = mail.outbox[1]
        self.assertEqual(
            message.subject,
            "Confirm your email change request",
        )
        self.assertEqual(message.to, ["newemail@example.com"])
        self.assertRegex(
            message.body,
            # Matches the email change confirmation URL:
            # http://testserver/manage/change-email-confirm/eyJ1dWlkIjoiMDE5MjZiNGYtODQ0Zi03MjRmLWE2YjQtMWQxYWEyYTU5OTgwIiwicmVjaXBpZW50IjoiY3VycmVudF9lbWFpbCJ9:1sy5S2:R7m51osUdabcMuOGXZRq7MabESIqKGl_mX2jO-TAcj8/
            r"http://testserver/manage/change-email-confirm/[0-9A-Za-z]+:[0-9a-zA-Z]+:[0-9A-Za-z_-]+/",
        )
        self.assertIn(
            "http://testserver/manage/change-email-cancel/",
            message.body,
        )

    def test_email_change_proposed_email_exists(self):
        """
        Pretend that the email change request was successful.

        When the user requests an email change to an email address that already exists,
        we pretend that the email change request was successful and send an email to the
        current email address, but a different one to the proposed one.
        """
        user_factories.UserFactory(email="existing@example.com")
        with (
            self.assertTemplateUsed("hidp/accounts/management/email/email_change_subject.txt"),
            self.assertTemplateUsed("hidp/accounts/management/email/email_change_body.txt"),
            self.assertTemplateUsed("hidp/accounts/management/email/email_change_body.html"),
            self.assertTemplateUsed("hidp/accounts/management/email/proposed_email_exists_subject.txt"),
            self.assertTemplateUsed("hidp/accounts/management/email/proposed_email_exists_body.txt"),
            self.assertTemplateUsed("hidp/accounts/management/email/proposed_email_exists_body.html"),
        ):  # fmt: skip
            response = self.client.post(
                self.url,
                {
                    "password": "P@ssw0rd!",
                    "proposed_email": "existing@example.com",
                },
                follow=True,
            )

        self.assertRedirects(
            response, reverse("hidp_account_management:email_change_request_sent")
        )
        self.assertTemplateUsed(
            response, "hidp/accounts/management/email_change_request_sent.html"
        )

        # EmailChangeRequest should be created
        self.assertTrue(
            EmailChangeRequest.objects.filter(
                user=self.user,
                proposed_email="existing@example.com",
            ).exists()
        )

        # Email should be sent to current email
        self.assertEqual(len(mail.outbox), 2)

        message = mail.outbox[0]
        self.assertEqual(
            message.subject,
            "Confirm your email change request",
        )
        self.assertEqual(message.to, [self.user.email])
        self.assertRegex(
            message.body,
            # Matches the email change confirmation URL:
            # http://testserver/manage/change-email-confirm/eyJ1dWlkIjoiMDE5MjZiNGYtODQ0Zi03MjRmLWE2YjQtMWQxYWEyYTU5OTgwIiwicmVjaXBpZW50IjoiY3VycmVudF9lbWFpbCJ9:1sy5S2:R7m51osUdabcMuOGXZRq7MabESIqKGl_mX2jO-TAcj8/
            r"http://testserver/manage/change-email-confirm/[0-9A-Za-z]+:[0-9a-zA-Z]+:[0-9A-Za-z_-]+/",
        )
        self.assertIn(
            "http://testserver/manage/change-email-cancel/",
            message.body,
        )

        # A different email should be sent to proposed email
        message = mail.outbox[1]
        self.assertEqual(
            message.subject,
            "Email change request",
        )
        self.assertEqual(message.to, ["existing@example.com"])
        self.assertIn(
            "However, you already have an account that uses existing@example.com",
            message.body,
        )
        self.assertIn(
            "http://testserver/manage/change-email-cancel/",
            message.body,
        )

    def test_proposed_email_user_inactive(self):
        inactive_user = user_factories.UserFactory(is_active=False)

        response = self.client.post(
            self.url,
            {
                "password": "P@ssw0rd!",
                "proposed_email": inactive_user.email,
            },
            follow=True,
        )

        self.assertRedirects(
            response, reverse("hidp_account_management:email_change_request_sent")
        )

        # Email should be sent to current email, but not to proposed email (inactive)
        self.assertEqual(len(mail.outbox), 1)

        message = mail.outbox[0]
        self.assertEqual(
            message.subject,
            "Confirm your email change request",
        )
        self.assertEqual(message.to, [self.user.email])


class TestEmailChangeRequestForm(TestCase):
    @classmethod
    def setUpTestData(cls):
        cls.user = user_factories.UserFactory()
        cls.existing_email_change_request = user_factories.EmailChangeRequestFactory(
            user=cls.user
        )

    def test_form_invalid_password(self):
        form = EmailChangeRequestForm(
            user=self.user,
            data={
                "password": "invalid",
                "proposed_email": "newemail@example.com",
            },
        )
        self.assertFalse(form.is_valid())

    def test_form_current_email(self):
        form = EmailChangeRequestForm(
            user=self.user,
            data={
                "password": "P@ssw0rd!",
                "proposed_email": self.user.email,
            },
        )
        self.assertFalse(form.is_valid(), msg="Expected form to be invalid")
        self.assertIn(
            "The new email address is the same as the current email address.",
            form.errors["proposed_email"],
        )

    def test_form_valid(self):
        form = EmailChangeRequestForm(
            user=self.user,
            data={
                "password": "P@ssw0rd!",
                "proposed_email": "newemail@example.com",
            },
        )
        self.assertTrue(form.is_valid())

        form.save()

        # Old EmailChangeRequest should be deleted
        self.assertFalse(
            EmailChangeRequest.objects.filter(
                id=self.existing_email_change_request.pk
            ).exists()
        )

        # New EmailChangeRequest should be created
        self.assertTrue(
            EmailChangeRequest.objects.filter(
                user=self.user,
                proposed_email="newemail@example.com",
            ).exists()
        )


class TestEmailChangeConfirm(TestCase):
    @classmethod
    def setUpTestData(cls):
        cls.user = user_factories.UserFactory(email="current@example.com")
        cls.email_change_request = user_factories.EmailChangeRequestFactory(
            user=cls.user, proposed_email="newemail@example.com"
        )
        cls.current_email_url = reverse(
            "hidp_account_management:email_change_confirm",
            kwargs={
                "token": tokens.email_change_token_generator.make_token(
                    str(cls.email_change_request.pk), "current_email"
                )
            },
        )
        cls.proposed_email_url = reverse(
            "hidp_account_management:email_change_confirm",
            kwargs={
                "token": tokens.email_change_token_generator.make_token(
                    str(cls.email_change_request.pk), "proposed_email"
                )
            },
        )

    def setUp(self):
        self.client.force_login(self.user)

    def _assert_response(self, response, *, validlink=True):
        """Convenience method to assert the response."""
        self.assertEqual(response.status_code, HTTPStatus.OK)
        if validlink:
            self.assertTemplateUsed(
                response, "hidp/accounts/management/email_change_confirm.html"
            )
        else:
            self.assertTemplateUsed(
                response,
                "hidp/accounts/management/email_change_confirm_invalid_link.html",
            )

    def test_get_unauthenticated_user(self):
        self.client.logout()
        response = self.client.get(self.current_email_url)

        self.assertEqual(response.status_code, HTTPStatus.FOUND)
        self.assertRedirects(
            response, f"{reverse('hidp_accounts:login')}?next={self.current_email_url}"
        )

    def test_get_invalid_token(self):
        self._assert_response(
            self.client.get(
                reverse(
                    "hidp_account_management:email_change_confirm",
                    kwargs={"token": "invalid"},
                ),
                follow=True,
            ),
            validlink=False,
        )

    def test_no_token_in_session(self):
        """Placeholder token, no token in session."""
        self._assert_response(
            self.client.get(
                reverse(
                    "hidp_account_management:email_change_confirm",
                    kwargs={"token": "email-change"},
                ),
                follow=True,
            ),
            validlink=False,
        )

    def test_valid_token_wrong_user(self):
        self.client.force_login(user_factories.UserFactory())

        self._assert_response(
            self.client.get(self.current_email_url, follow=True), validlink=False
        )

    def test_get_valid_token(self):
        self._assert_response(self.client.get(self.current_email_url, follow=True))

    def test_get_already_confirmed(self):
        self.email_change_request.confirmed_by_current_email = True
        self.email_change_request.save()

        self._assert_response(
            self.client.get(self.current_email_url, follow=True), validlink=False
        )

    def test_post_current_email_valid_token(self):
        response = self.client.post(
            self.current_email_url, {"allow_change": "on"}, follow=True
        )
        self.assertRedirects(
            response,
            reverse("hidp_account_management:email_change_complete"),
            status_code=307,
        )
        self.assertTemplateUsed(
            response, "hidp/accounts/management/email_change_complete.html"
        )
        self.assertInHTML(
            "Please also confirm the change from your new email address.",
            response.content.decode(),
        )
        self.email_change_request.refresh_from_db()
        self.assertEqual(self.email_change_request.confirmed_by_current_email, True)

        # Email address should not be changed yet
        self.user.refresh_from_db()
        self.assertEqual(self.user.email, "current@example.com")

        # Email changed mail should not be sent yet
        self.assertEqual(len(mail.outbox), 0)

    def test_post_proposed_email_valid_token(self):
        response = self.client.post(
            self.proposed_email_url, {"allow_change": "on"}, follow=True
        )
        self.assertRedirects(
            response,
            reverse("hidp_account_management:email_change_complete"),
            status_code=307,
        )
        self.assertTemplateUsed(
            response, "hidp/accounts/management/email_change_complete.html"
        )
        self.assertInHTML(
            "Please also confirm the change from your current email address.",
            response.content.decode(),
        )
        self.email_change_request.refresh_from_db()
        self.assertEqual(self.email_change_request.confirmed_by_proposed_email, True)

        # Email address should not be changed yet
        self.user.refresh_from_db()
        self.assertEqual(self.user.email, "current@example.com")

        # Email changed mail should not be sent yet
        self.assertEqual(len(mail.outbox), 0)

    def test_post_second_valid_token(self):
        self.email_change_request.confirmed_by_current_email = True
        self.email_change_request.save()

        with (
            self.assertTemplateUsed("hidp/accounts/management/email/email_changed_subject.txt"),
            self.assertTemplateUsed("hidp/accounts/management/email/email_changed_body.txt"),
            self.assertTemplateUsed("hidp/accounts/management/email/email_changed_body.html"),
        ):  # fmt: skip
            response = self.client.post(
                self.proposed_email_url, {"allow_change": "on"}, follow=True
            )

        self.assertRedirects(
            response,
            reverse("hidp_account_management:email_change_complete"),
            status_code=307,
        )
        self.assertTemplateUsed(
            response, "hidp/accounts/management/email_change_complete.html"
        )
        self.assertInHTML(
            "Your account email address has been changed successfully.",
            response.content.decode(),
        )

        self.user.refresh_from_db()
        self.assertEqual(self.user.email, "newemail@example.com")

        email_change_request = EmailChangeRequest.objects.filter(
            user=self.user, proposed_email="newemail@example.com"
        )
        self.assertTrue(email_change_request.exists())
        self.assertTrue(email_change_request.first().is_complete())

        # Email changed mail should be sent
        self.assertEqual(len(mail.outbox), 1)
        self.assertEqual(
            "Your account email address has been changed",
            mail.outbox[0].subject,
        )
        self.assertEqual(
            mail.outbox[0].to,
            ["current@example.com", "newemail@example.com"],
        )

    def test_post_invalid_token(self):
        self._assert_response(
            self.client.post(
                reverse(
                    "hidp_account_management:email_change_confirm",
                    kwargs={"token": "invalid"},
                ),
                {"allow_change": "on"},
                follow=True,
            ),
            validlink=False,
        )

    def test_post_proposed_email_already_exists(self):
        # Should only happen if an account was created with the proposed email
        # address after email change request was made.
        user_factories.UserFactory(email="newemail@example.com")
        self.email_change_request.confirmed_by_current_email = True
        self.email_change_request.save()

        response = self.client.post(
            self.proposed_email_url, {"allow_change": "on"}, follow=True
        )
        self.assertEqual(response.status_code, HTTPStatus.OK)
        self.assertFalse(
            response.context["form"].is_valid(), msg="Expected form to be invalid"
        )
        self.assertIn(
            "Sorry, changing your email address is not possible because an"
            " account with this email address already exists.",
            response.context["form"].errors["__all__"],
        )

    def test_post_already_completed_request(self):
        self.email_change_request.confirmed_by_current_email = True
        self.email_change_request.confirmed_by_proposed_email = True
        self.email_change_request.save()

        self._assert_response(
            self.client.post(
                self.current_email_url, {"allow_change": "on"}, follow=True
            ),
            validlink=False,
        )


class TestEmailChangeCancel(TestCase):
    @classmethod
    def setUpTestData(cls):
        cls.user = user_factories.UserFactory()
        cls.email_change_request = user_factories.EmailChangeRequestFactory(
            user=cls.user
        )
        cls.url = reverse("hidp_account_management:email_change_cancel")

    def setUp(self):
        self.client.force_login(self.user)

    def test_get_unauthenticated_user(self):
        self.client.logout()
        response = self.client.get(self.url)

        self.assertEqual(response.status_code, HTTPStatus.FOUND)
        self.assertRedirects(
            response, f"{reverse('hidp_accounts:login')}?next={self.url}"
        )

    def _assert_response(self, response, *, validlink=True):
        """Convenience method to assert the response."""
        self.assertEqual(response.status_code, HTTPStatus.OK)
        if validlink:
            self.assertTemplateUsed(
                response, "hidp/accounts/management/email_change_cancel.html"
            )
        else:
            self.assertTemplateUsed(
                response,
                "hidp/accounts/management/email_change_cancel_invalid_link.html",
            )

    def test_get_no_change_request(self):
        self.email_change_request.delete()

        self._assert_response(self.client.get(self.url), validlink=False)

    def test_get_completed_change_request(self):
        self.email_change_request.confirmed_by_current_email = True
        self.email_change_request.confirmed_by_proposed_email = True
        self.email_change_request.save()

        self._assert_response(self.client.get(self.url), validlink=False)

    def test_get(self):
        self._assert_response(self.client.get(self.url))

    def test_post_expired_change_request(self):
        self.email_change_request.created_at = timezone.now() - timezone.timedelta(
            days=8
        )
        self.email_change_request.save()

        self._assert_response(
            self.client.post(self.url, {"allow_cancel": "on"}), validlink=False
        )

    def test_post_no_change_request(self):
        self.email_change_request.delete()

        self._assert_response(
            self.client.post(self.url, {"allow_cancel": "on"}), validlink=False
        )

    def test_post(self):
        response = self.client.post(self.url, {"allow_cancel": "on"}, follow=True)
        self.assertRedirects(
            response,
            reverse("hidp_account_management:email_change_cancel_done"),
        )
        self.assertTemplateUsed(
            response, "hidp/accounts/management/email_change_cancel_done.html"
        )

        self.assertFalse(
            EmailChangeRequest.objects.filter(
                user=self.user,
            ).exists()
        )


class TestRemoveIncompleteEmailChangeRequests(TestCase):
    @classmethod
    def setUpTestData(cls):
        cls.stale_request = user_factories.EmailChangeRequestFactory()
        cls.stale_request.created_at = timezone.now() - timezone.timedelta(days=14)
        cls.stale_request.save()

        cls.incomplete_recent_request = user_factories.EmailChangeRequestFactory()
        cls.incomplete_recent_request.created_at = timezone.now() - timezone.timedelta(
            days=3
        )
        cls.incomplete_recent_request.save()

        cls.completed_request = user_factories.EmailChangeRequestFactory(
            confirmed_by_current_email=True,
            confirmed_by_proposed_email=True,
        )
        cls.completed_request.created_at = timezone.now() - timezone.timedelta(days=1)
        cls.completed_request.save()

    def test_remove_complete_and_stale_email_change_requests_dry_run(self):
        removed_requests = remove_complete_and_stale_email_change_requests(dry_run=True)
        email_change_requests_exist = (
            EmailChangeRequest.objects.filter(pk=email_change_request.pk).exists()
            for email_change_request in (
                self.stale_request,
                self.incomplete_recent_request,
                self.completed_request,
            )
        )
        self.assertEqual(
            removed_requests,
            2,
            msg="Expected 2 requests to be selected for removal.",
        )
        self.assertTrue(
            all(email_change_requests_exist),
            msg="Expected all requests to still exist after dry run.",
        )

    def test_remove_complete_and_stale_email_change_requests(self):
        removed_requests = remove_complete_and_stale_email_change_requests()
        self.assertEqual(
            removed_requests,
            2,
            msg="Expected 2 requests to be selected for removal.",
        )
        self.assertFalse(
            EmailChangeRequest.objects.filter(pk=self.stale_request.pk).exists(),
            msg="Expected stale request to be removed, created more than 7 days ago.",
        )
        self.assertFalse(
            EmailChangeRequest.objects.filter(pk=self.completed_request.pk).exists(),
            msg="Expected complete request to be removed.",
        )
        self.assertTrue(
            EmailChangeRequest.objects.filter(
                pk=self.incomplete_recent_request.pk
            ).exists(),
            msg=(
                "Expected incomplete recent request to still exist, created less"
                " than 7 days ago.",
            ),
        )

    def test_remove_complete_and_stale_email_change_requests_2_days(self):
        removed_requests = remove_complete_and_stale_email_change_requests(days=2)
        self.assertEqual(
            removed_requests,
            3,
            msg="Expected 3 requests to be selected for removal.",
        )
        self.assertFalse(
            EmailChangeRequest.objects.filter(pk=self.stale_request.pk).exists(),
            msg="Expected stale request to be removed, created more than 2 days ago.",
        )
        self.assertFalse(
            EmailChangeRequest.objects.filter(pk=self.completed_request.pk).exists(),
            msg="Expected complete request to be removed.",
        )
        self.assertFalse(
            EmailChangeRequest.objects.filter(
                pk=self.incomplete_recent_request.pk
            ).exists(),
            msg=(
                "Expected incomplete recent request to be removed, created more"
                " than 2 days ago.",
            ),
        )

    @mock.patch(
        "hidp.accounts.management.commands.remove_complete_and_stale_email_change_requests.remove_complete_and_stale_email_change_requests",
        return_value=1,
    )
    def test_remove_complete_and_stale_email_change_requests_management_command_dry_run(
        self, mock_remove_complete_and_stale_email_change_requests
    ):
        stdout = io.StringIO()

        call_command(
            "remove_complete_and_stale_email_change_requests",
            dry_run=True,
            stdout=stdout,
        )
        self.assertIn(
            "Removing completed email change requests and requests that have not been"
            " completed within 7 days...",
            stdout.getvalue(),
        )
        self.assertIn(
            "1 completed and/or stale email change request(s) would be removed.",
            stdout.getvalue(),
        )
        mock_remove_complete_and_stale_email_change_requests.assert_called_once_with(
            days=7, dry_run=True
        )

    @mock.patch(
        "hidp.accounts.management.commands.remove_complete_and_stale_email_change_requests.remove_complete_and_stale_email_change_requests",
        return_value=1,
    )
    def test_remove_complete_and_stale_email_change_requests_management_command(
        self, mock_remove_complete_and_stale_email_change_requests
    ):
        stdout = io.StringIO()

        call_command("remove_complete_and_stale_email_change_requests", stdout=stdout)
        self.assertIn(
            "Removing completed email change requests and requests that have not been"
            " completed within 7 days...",
            stdout.getvalue(),
        )
        self.assertIn(
            "Successfully removed 1 completed and/or stale email change request(s).",
            stdout.getvalue(),
        )
        mock_remove_complete_and_stale_email_change_requests.assert_called_once_with(
            days=7, dry_run=False
        )

    @mock.patch(
        "hidp.accounts.management.commands.remove_complete_and_stale_email_change_requests.remove_complete_and_stale_email_change_requests",
        return_value=2,
    )
    def test_remove_complete_and_stale_email_change_requests_management_command_2_days(
        self, mock_remove_complete_and_stale_email_change_requests
    ):
        stdout = io.StringIO()

        call_command(
            "remove_complete_and_stale_email_change_requests", days=2, stdout=stdout
        )
        self.assertIn(
            "Removing completed email change requests and requests that have not been"
            " completed within 2 days...",
            stdout.getvalue(),
        )
        self.assertIn(
            "Successfully removed 2 completed and/or stale email change request(s).",
            stdout.getvalue(),
        )
        mock_remove_complete_and_stale_email_change_requests.assert_called_once_with(
            days=2, dry_run=False
        )



import factory

from factory.django import DjangoModelFactory

from django.conf import settings
from django.utils import timezone


class UserFactory(DjangoModelFactory):
    class Meta:
        model = settings.AUTH_USER_MODEL
        skip_postgeneration_save = True

    first_name = factory.Faker("first_name")
    last_name = factory.Faker("last_name")
    email = factory.Faker("email")
    password = factory.django.Password("P@ssw0rd!")


class VerifiedUserFactory(UserFactory):
    email_verified = factory.LazyFunction(timezone.now)


class SuperUserFactory(VerifiedUserFactory):
    is_superuser = True
    is_staff = True


class EmailChangeRequestFactory(DjangoModelFactory):
    class Meta:
        model = "hidp_accounts.EmailChangeRequest"

    user = factory.SubFactory(VerifiedUserFactory)
    proposed_email = factory.Faker("email")
    current_email = factory.SelfAttribute("user.email")



from http import HTTPStatus
from unittest import mock

from django.contrib.auth import get_user_model
from django.core import mail
from django.test import TransactionTestCase, override_settings
from django.urls import reverse
from django.utils import timezone

from hidp.accounts.forms import UserCreationForm
from hidp.test.factories import user_factories

User = get_user_model()


@override_settings(LANGUAGE_CODE="en", REGISTRATION_ENABLED=True)
class TestRegistrationView(TransactionTestCase):
    def setUp(self):
        self.test_user = user_factories.UserFactory(email="user@example.com")
        self.signup_url = reverse("hidp_accounts:register")

    def test_get(self):
        """The registration form should be displayed."""
        response = self.client.get(self.signup_url)
        self.assertEqual(response.status_code, 200)
        self.assertTemplateUsed(response, "hidp/accounts/register.html")
        self.assertIn("form", response.context)
        self.assertIsInstance(response.context["form"], UserCreationForm)

    def test_get_tos(self):
        """The terms of service should be displayed."""
        response = self.client.get(reverse("hidp_accounts:tos"))
        self.assertEqual(response.status_code, 200)
        self.assertTemplateUsed(response, "hidp/accounts/tos.html")

    def test_tos_required(self):
        """The user should agree to the terms of service."""
        response = self.client.post(
            self.signup_url,
            {
                "email": "test@example.com",
                "password1": "P@ssw0rd!",
                "password2": "P@ssw0rd!",
            },
        )
        self.assertFormError(
            response.context["form"], "agreed_to_tos", "This field is required."
        )

    def test_valid_registration(self):
        """A new user should be created and logged in."""
        with (
            self.assertTemplateUsed("hidp/accounts/verification/email/verification_subject.txt"),
            self.assertTemplateUsed("hidp/accounts/verification/email/verification_body.txt"),
            self.assertTemplateUsed("hidp/accounts/verification/email/verification_body.html"),
        ):  # fmt: skip
            response = self.client.post(
                self.signup_url,
                {
                    "email": "test@example.com",
                    "password1": "P@ssw0rd!",
                    "password2": "P@ssw0rd!",
                    "agreed_to_tos": "on",
                },
                follow=True,
            )
        self.assertTrue(
            User.objects.filter(email="test@example.com").exists(),
            msg="Expected user to be created",
        )
        user = User.objects.get(email="test@example.com")
        # Agreed to TOS
        self.assertAlmostEqual(
            timezone.now(),
            user.agreed_to_tos,
            delta=timezone.timedelta(seconds=10),
        )
        # Verification email sent
        self.assertEqual(len(mail.outbox), 1)
        message = mail.outbox[0]
        self.assertEqual(
            message.subject,
            "Verify your email address",
        )
        # Redirected to verification required page
        self.assertRedirects(
            response,
            reverse(
                "hidp_accounts:email_verification_required", kwargs={"token": "email"}
            ),
        )
        # Verification required page
        self.assertInHTML(
            "Verification required",
            response.content.decode("utf-8"),
        )

    def test_valid_registration_safe_next_param(self):
        response = self.client.post(
            self.signup_url,
            {
                "email": "test@example.com",
                "password1": "P@ssw0rd!",
                "password2": "P@ssw0rd!",
                "agreed_to_tos": "on",
                "next": "/example/",
            },
            follow=True,
        )
        self.assertTrue(
            User.objects.filter(email="test@example.com").exists(),
            msg="Expected user to be created",
        )
        # Redirected to verification required page
        self.assertRedirects(
            response,
            reverse(
                "hidp_accounts:email_verification_required", kwargs={"token": "email"}
            )
            + "?next=/example/",
        )

    def test_valid_registration_unsafe_next_param(self):
        response = self.client.post(
            self.signup_url,
            {
                "email": "test@example.com",
                "password1": "P@ssw0rd!",
                "password2": "P@ssw0rd!",
                "agreed_to_tos": "on",
                "next": "https://example.com/",
            },
            follow=True,
        )
        self.assertTrue(
            User.objects.filter(email="test@example.com").exists(),
            msg="Expected user to be created",
        )
        # Redirected to verification required page
        self.assertRedirects(
            response,
            reverse(
                "hidp_accounts:email_verification_required", kwargs={"token": "email"}
            ),
        )

    def test_duplicate_email_unverified(self):
        """Signup using an exiting email should look like a successful signup."""
        response = self.client.post(
            self.signup_url,
            {
                # Different case, still considered duplicate
                "email": "USER@EXAMPLE.COM",
                "password1": "P@ssw0rd!",
                "password2": "P@ssw0rd!",
                "agreed_to_tos": "on",
            },
            follow=True,
        )
        # Redirected to verification required page
        self.assertRedirects(
            response,
            reverse(
                "hidp_accounts:email_verification_required", kwargs={"token": "email"}
            ),
        )
        # Verification required page
        self.assertInHTML(
            "Verification required",
            response.content.decode("utf-8"),
        )

    def test_duplicate_email_verified(self):
        """Verified users should get a reminder mail."""
        self.test_user.email_verified = timezone.now()
        self.test_user.save()
        with (
            self.assertTemplateUsed("hidp/accounts/verification/email/account_exists_subject.txt"),
            self.assertTemplateUsed("hidp/accounts/verification/email/account_exists_body.txt"),
            self.assertTemplateUsed("hidp/accounts/verification/email/account_exists_body.html"),
        ):  # fmt: skip
            response = self.client.post(
                self.signup_url,
                {
                    # Different case, still considered duplicate
                    "email": "USER@EXAMPLE.COM",
                    "password1": "P@ssw0rd!",
                    "password2": "P@ssw0rd!",
                    "agreed_to_tos": "on",
                },
                follow=True,
            )
        # Redirected to verification required page
        self.assertRedirects(
            response,
            reverse(
                "hidp_accounts:email_verification_required", kwargs={"token": "email"}
            ),
        )
        # Verification required page
        self.assertInHTML(
            "Verification required",
            response.content.decode("utf-8"),
        )
        # Sends an email notification to the user
        self.assertEqual(len(mail.outbox), 1)
        message = mail.outbox[0]
        self.assertEqual(message.to, [self.test_user.email])
        self.assertEqual("Sign up request", message.subject)

    def test_with_logged_in_user(self):
        """A logged-in user should not be able to sign up again."""
        self.client.force_login(self.test_user)
        response = self.client.post(
            self.signup_url,
            {
                "email": "test@example.com",
                "password1": "P@ssw0rd!",
                "password2": "P@ssw0rd!",
                "agreed_to_tos": "on",
            },
        )
        self.assertEqual(response.status_code, HTTPStatus.FORBIDDEN)

    def test_account_exists(self):
        """Existing users should be notified if someone tries to use their email."""
        existing_user = user_factories.VerifiedUserFactory()
        with (
            self.assertTemplateUsed("hidp/accounts/verification/email/account_exists_subject.txt"),
            self.assertTemplateUsed("hidp/accounts/verification/email/account_exists_body.txt"),
            self.assertTemplateUsed("hidp/accounts/verification/email/account_exists_body.html"),
        ):  # fmt: skip
            response = self.client.post(
                self.signup_url,
                {
                    "email": existing_user.email,
                    "password1": "P@ssw0rd!",
                    "password2": "P@ssw0rd!",
                    "agreed_to_tos": "on",
                },
                follow=True,
            )

        # Pretend the registration was successful
        self.assertRedirects(
            response,
            reverse(
                "hidp_accounts:email_verification_required",
                kwargs={"token": "email"},
            ),
        )

        # Account exists email sent
        self.assertEqual(len(mail.outbox), 1)
        message = mail.outbox[0]
        self.assertEqual(
            message.subject,
            "Sign up request",
        )
        self.assertIn(
            "You're receiving this email because you attempted to create an"
            " account using this email address."
            " However, an account already exists with this email address.",
            message.body,
        )

    def test_inactive_account_exists(self):
        """Inactive users should not be notified if someone tries to use their email."""
        inactive_user = user_factories.VerifiedUserFactory(is_active=False)
        response = self.client.post(
            self.signup_url,
            {
                "email": inactive_user.email,
                "password1": "P@ssw0rd!",
                "password2": "P@ssw0rd!",
                "agreed_to_tos": "on",
            },
            follow=True,
        )

        # Pretend the registration was successful
        self.assertRedirects(
            response,
            reverse(
                "hidp_accounts:email_verification_required",
                kwargs={"token": "email"},
            ),
        )

        # Account exists email not sent
        self.assertEqual(len(mail.outbox), 0)

    @mock.patch(
        "hidp.accounts.views.mailers.EmailVerificationMailer.send",
        side_effect=Exception,
    )
    def test_verification_email_error(self, mock_send):
        """Errors are logged when sending the verification email."""
        # This is a fix for Django's CVE-2024-45231. The email backend
        # might raise an exception when sending an email, which could
        # be used to enumerate valid email addresses.
        with self.assertLogs("hidp.accounts.views", level="ERROR") as cm:
            self.client.post(
                self.signup_url,
                {
                    "email": "test@example.com",
                    "password1": "P@ssw0rd!",
                    "password2": "P@ssw0rd!",
                    "agreed_to_tos": "on",
                },
            )
        self.assertIn("Failed to send verification email.", cm.records[0].msg)
        self.assertEqual(0, len(mail.outbox))

    @mock.patch(
        "hidp.accounts.views.mailers.AccountExistsMailer.send",
        side_effect=Exception,
    )
    def test_account_exists_email_error(self, mock_send):
        """Errors are logged when sending the account exists email."""
        # This is a fix for Django's CVE-2024-45231. The email backend
        # might raise an exception when sending an email, which could
        # be used to enumerate valid email addresses.
        user_factories.VerifiedUserFactory(email="test@example.com")
        with self.assertLogs("hidp.accounts.views", level="ERROR") as cm:
            self.client.post(
                self.signup_url,
                {
                    "email": "test@example.com",
                    "password1": "P@ssw0rd!",
                    "password2": "P@ssw0rd!",
                    "agreed_to_tos": "on",
                },
            )
        self.assertIn("Failed to send verification email.", cm.records[0].msg)
        self.assertEqual(0, len(mail.outbox))


@override_settings(REGISTRATION_ENABLED=False)
class TestRegistrationViewNotEnabled(TransactionTestCase):
    def setUp(self):
        self.signup_url = reverse("hidp_accounts:register")

    def test_get_signup_disabled(self):
        """The view should turn a 404 when registration is disabled."""
        response = self.client.get(self.signup_url)
        self.assertEqual(response.status_code, HTTPStatus.NOT_FOUND)

    def test_post_signup_disabled(self):
        """The view should not process POST requests when registration is disabled."""
        response = self.client.post(
            self.signup_url,
            {
                "email": "test@example.com",
                "password1": "P@ssw0rd!",
                "password2": "P@ssw0rd!",
                "agreed_to_tos": "on",
            },
        )
        self.assertEqual(response.status_code, HTTPStatus.NOT_FOUND)



from django.test import RequestFactory, TestCase

from hidp.rate_limit.decorators import rate_limit_default, rate_limit_strict


@rate_limit_default
def rate_limit_default_view(request):
    pass


@rate_limit_strict
def rate_limit_strict_view(request):
    pass


class TestRateLimitDecorators(TestCase):
    def test_rate_limit_default(self):
        request = RequestFactory().get("/")
        rate_limit_default_view(request)
        self.assertIsNotNone(getattr(request, "limited", None))

        request = RequestFactory().post("/")
        rate_limit_default_view(request)
        self.assertIsNotNone(getattr(request, "limited", None))

    def test_rate_limit_strict(self):
        request = RequestFactory().get("/")
        rate_limit_strict_view(request)
        self.assertIsNotNone(getattr(request, "limited", None))

        request = RequestFactory().post("/")
        rate_limit_strict_view(request)
        self.assertIsNotNone(getattr(request, "limited", None))



from datetime import UTC
from importlib import import_module

from oauth2_provider import oauth2_validators
from oauth2_provider.models import get_access_token_model

from django.conf import settings
from django.contrib import auth
from django.contrib.auth.models import AnonymousUser
from django.http import HttpRequest, parse_cookie
from django.utils import timezone
from django.utils.timezone import localtime

AccessTokenModel = get_access_token_model()


def _get_user_from_oauthlib_request(request):
    # Request is an oauthlib request, not a Django request, so the session
    # and user need to be determined from the headers.
    cookies = parse_cookie(request.headers["HTTP_COOKIE"])
    session_id = cookies.get(settings.SESSION_COOKIE_NAME)
    if not session_id:
        # Session cookie not found, no user to authenticate.
        return AnonymousUser()
    # auth.get_user requires a Django request, with a session object,
    # in order to look up the user. Create a dummy request object and
    # load the session identified by the session cookie.
    session_engine = import_module(settings.SESSION_ENGINE)
    request = HttpRequest()
    request.session = session_engine.SessionStore(session_id)
    # Find the user associated with the session and return it.
    # If no user is logged in AnonymousUser will be returned.
    return auth.get_user(request)


class OAuth2Validator(oauth2_validators.OAuth2Validator):
    def validate_silent_login(self, request):  # noqa: PLR6301 (no-self-use)
        """
        Check if the user is logged in to determine if silent login is possible.

        This is further validated in `validate_silent_authorization`.
        """
        # Assign the user to the oauthlib request object.
        request.user = _get_user_from_oauthlib_request(request)
        # Silently authenticate the user if the user is logged in.
        return request.user.is_authenticated

    def validate_silent_authorization(self, request):  # noqa: PLR6301 (no-self-use)
        """
        Determine if the logged-in user has authorized the application.

        If so, silent authorization is possible.
        """
        # validate_silent_login is called before validate_silent_authorization,
        # so the user *should* be assigned to the request object.
        user = getattr(request, "user", None)
        if not user or not user.is_authenticated:
            return False

        if request.client.skip_authorization:
            # The client has been configured to skip the authorization step.
            return True

        # DOT does not record whether a user has consented to an application.
        # It assumes that, if an active access token exists, the user has consented.
        # Once the access tokens expire, the user will need to re-consent.
        tokens = AccessTokenModel.objects.filter(
            user=request.user, application=request.client, expires__gte=timezone.now()
        ).all()
        return any(token.allow_scopes(request.scopes) for token in tokens)

    # Maps OIDC claims to scopes. This is used to determine which claims to include
    # in the ID token and user info response.
    oidc_claim_scope = oauth2_validators.OAuth2Validator.oidc_claim_scope

    def get_additional_claims(self, request):
        """
        Map user attributes to OIDC claims.

        These claims are included in the ID token and user info response.

        Only those claims that belong to the requested scopes are included.
        This means that, for example, if the client only requests the `openid` scope,
        none of the additional claims will be included in the ID token.

        The mapping of claims to scopes is defined in the `oidc_claim_scope` dictionary.
        """
        return super().get_additional_claims(request) | {
            "name": request.user.get_full_name(),
            "given_name": request.user.first_name,
            "family_name": request.user.last_name,
            "updated_at": int(localtime(request.user.last_modified, UTC).timestamp()),
            "email": request.user.email,
            "email_verified": request.user.email_verified is not None,
        }



import importlib

from django.conf import settings
from django.contrib.auth import get_user_model
from django.core import checks
from django.urls import NoReverseMatch, reverse

from ..accounts.models import BaseUser

REQUIRED_APPS = [
    "django.contrib.contenttypes",
    "django.contrib.auth",
    "django.contrib.sessions",
    "hidp",
    "hidp.accounts",
    "hidp.csp",
    "hidp.federated",
]

OIDC_PROVIDER_REQUIRED_APPS = [
    "oauth2_provider",
    "hidp.oidc_provider",
    "rest_framework",
    "hidp.api",
]

OTP_REQUIRED_APPS = [
    "django_otp",
    "django_otp.plugins.otp_totp",
    "django_otp.plugins.otp_static",
    "hidp.otp",
]

REQUIRED_MIDDLEWARE = [
    "django.contrib.sessions.middleware.SessionMiddleware",
    "django.middleware.csrf.CsrfViewMiddleware",
    "django.contrib.auth.middleware.AuthenticationMiddleware",
    "hidp.rate_limit.middleware.RateLimitMiddleware",
]

OTP_REQUIRED_MIDDLEWARE = "django_otp.middleware.OTPMiddleware"


class Tags:
    dependencies = "dependencies"
    middleware = "middleware"
    settings = "settings"


# Make sure the required apps are installed
E001 = checks.Error(
    "INSTALLED_APPS does not include the required apps for HIdP to work.",
    hint="INSTALLED_APPS should include the following apps: {}.".format(
        ", ".join(f"{app_name!r}" for app_name in REQUIRED_APPS)
    ),
    id="hidp.E001",
)


@checks.register(Tags.dependencies)
def check_installed_apps(**kwargs):
    for app_name in REQUIRED_APPS:
        if app_name not in settings.INSTALLED_APPS:
            return [E001]
    return []


# Make sure the required middleware is included
E002 = checks.Error(
    "MIDDLEWARE does not include the required middleware for HIdP to work.",
    hint="MIDDLEWARE should include the following middleware: {}.".format(
        ", ".join(f"{middleware!r}" for middleware in REQUIRED_MIDDLEWARE)
    ),
    id="hidp.E002",
)


@checks.register(Tags.middleware)
def check_middleware(**kwargs):
    for middleware in REQUIRED_MIDDLEWARE:
        if middleware not in settings.MIDDLEWARE:
            return [E002]
    return []


# Storing timezone-aware datetime objects is crucial for correctly handling
# timestamps in tokens and other time-sensitive operations.
E003 = checks.Error(
    "USE_TZ is not set to True.",
    hint="Set USE_TZ to True.",
    id="hidp.E003",
)


@checks.register(Tags.settings)
def check_use_tz(**kwargs):
    if not settings.USE_TZ:
        return [E003]
    return []


# Make sure the user model is compatible with HIdP
E004 = checks.Error(
    "AUTH_USER_MODEL is not set to a subclass of hidp.accounts.models.BaseUser.",
    hint="Define your own subclass of BaseUser and set AUTH_USER_MODEL.",
    id="hidp.E004",
)


@checks.register(Tags.settings)
def check_user_model(**kwargs):
    user_model = get_user_model()
    if not issubclass(user_model, BaseUser):
        return [E004]
    return []


# Make sure Django OAuth Toolkit is configured
E005 = checks.Error(
    "OAUTH2_PROVIDER is not configured correctly.",
    hint="Use hidp.config.get_oauth2_provider_settings() to configure OAUTH2_PROVIDER.",
    id="hidp.E005",
)


def check_oauth2_provider(**kwargs):
    oauth2_provider_settings = getattr(settings, "OAUTH2_PROVIDER", None)
    if (
        oauth2_provider_settings is None
        or not isinstance(oauth2_provider_settings, dict)
        or "OIDC_RSA_PRIVATE_KEY" not in oauth2_provider_settings
    ):
        return [E005]
    return []


# Make sure the urls are configured correctly
E006 = checks.Error(
    "Unable to reverse the 'hidp_accounts:login' URL.",
    hint=(
        "Include hidp.config.urls in your ROOT_URLCONF,"
        " or define custom URLs using the 'hidp_accounts' namespace."
    ),
    id="hidp.E006",
)


@checks.register(Tags.settings)
def check_login_url(**kwargs):
    try:
        reverse("hidp_accounts:login")
    except NoReverseMatch:
        return [E006]
    return []


E008 = checks.Error(
    "INSTALLED_APPS does not include the required OIDC provider apps.",
    hint="INSTALLED_APPS should include the following apps: {}.".format(
        ", ".join(f"{app_name!r}" for app_name in OIDC_PROVIDER_REQUIRED_APPS)
    ),
    id="hidp.E008",
)


def check_oidc_provider_installed_apps(**kwargs):
    for app_name in OIDC_PROVIDER_REQUIRED_APPS:
        if app_name not in settings.INSTALLED_APPS:
            return [E008]
    return []


if importlib.util.find_spec("oauth2_provider") is not None:
    # Only enable the OIDC provider checks if OAuth2 Provider is installed
    checks.register(Tags.settings)(check_oauth2_provider)
    checks.register(Tags.dependencies)(check_oidc_provider_installed_apps)


# Make sure the required apps for OTP are installed
E009 = checks.Error(
    "INSTALLED_APPS does not include the required OTP apps.",
    hint="INSTALLED_APPS should include the following apps: {}.".format(
        ", ".join(f"{app_name!r}" for app_name in OTP_REQUIRED_APPS)
    ),
    id="hidp.E009",
)


@checks.register(Tags.dependencies)
def check_otp_installed_apps(**kwargs):
    if "hidp.otp" in settings.INSTALLED_APPS:
        for app_name in OTP_REQUIRED_APPS:
            if app_name not in settings.INSTALLED_APPS:
                return [E009]
    return []


# Make sure the required middleware for OTP is included
E010 = checks.Error(
    "MIDDLEWARE does not include the required middleware for OTP to work.",
    hint=(
        f'Add "{OTP_REQUIRED_MIDDLEWARE}" middleware after "AuthenticationMiddleware".'
    ),
    id="hidp.E010",
)


@checks.register(Tags.middleware)
def check_otp_middleware(**kwargs):
    if (
        "hidp.otp" in settings.INSTALLED_APPS
        and OTP_REQUIRED_MIDDLEWARE not in settings.MIDDLEWARE
    ):
        return [E010]
    return []


# If django-otp is installed but hidp.otp is not, show a warning
W001 = checks.Warning(
    "django-otp is installed but hidp.otp is not in INSTALLED_APPS.",
    hint="Consider adding 'hidp.otp' to INSTALLED_APPS for a more complete OTP"
    " implementation.",
    id="hidp.W001",
)


@checks.register(Tags.dependencies)
def check_hidp_otp_installed_apps_when_django_otp_installed(**kwargs):
    if (
        importlib.util.find_spec("django_otp") is not None
        and "hidp.otp" not in settings.INSTALLED_APPS
    ):
        return [W001]
    return []



from urllib.parse import urljoin

from django.contrib.auth.tokens import default_token_generator
from django.core.mail import EmailMultiAlternatives
from django.template import loader
from django.urls import reverse
from django.utils.encoding import force_bytes
from django.utils.http import urlsafe_base64_encode

from . import email_verification, tokens
from .email_change import Recipient


class BaseMailer:
    """Base class for sending templated emails."""

    subject_template_name = NotImplemented
    email_template_name = NotImplemented
    html_email_template_name = None

    def __init__(self, *, base_url):
        """
        Initialize the mailer.

        Args:
            base_url:
                The base URL to use when generating links in the email.
        """
        self.base_url = base_url.rstrip("/")

    def get_context(self, extra_context=None):
        """Return a dictionary of context variables for the email templates."""
        context = {
            "base_url": self.base_url,
        }
        return context | (extra_context or {})

    def get_recipients(self):
        """Return a list of email addresses to send the email to."""
        raise NotImplementedError("Method get_recipients must be implemented")

    def _get_subject(self, context):
        if self.subject_template_name is NotImplemented:
            raise NotImplementedError("Attribute subject_template_name must be set")
        subject = loader.render_to_string(self.subject_template_name, context)
        # Email subject *must not* contain newlines
        return "".join(subject.splitlines())

    def _get_body(self, context):
        if self.email_template_name is NotImplemented:
            raise NotImplementedError("Attribute email_template_name must be set")
        return loader.render_to_string(self.email_template_name, context)

    def _add_optional_html_body(self, email_message, context):
        if self.html_email_template_name is not None:
            html_email = loader.render_to_string(self.html_email_template_name, context)
            email_message.attach_alternative(html_email, "text/html")

    def _get_message(self, from_email=None, extra_context=None):
        context = self.get_context(extra_context)
        subject = self._get_subject(context)
        body = self._get_body(context)

        email_message = EmailMultiAlternatives(
            subject, body, from_email, self.get_recipients()
        )
        self._add_optional_html_body(email_message, context)

        return email_message

    def send(
        self,
        *,
        from_email=None,
        extra_context=None,
    ):
        """
        Send a django.core.mail.EmailMultiAlternatives.

        Args:
            from_email:
                Email address to use as the sender of the email.
                Optional, uses the `DEFAULT_FROM_EMAIL` setting if not provided.
            extra_context:
                A dictionary of extra context variables to use when rendering the
                email templates (optional).
        """
        self._get_message(from_email, extra_context).send()


class EmailVerificationMailer(BaseMailer):
    subject_template_name = "hidp/accounts/verification/email/verification_subject.txt"
    email_template_name = "hidp/accounts/verification/email/verification_body.txt"
    html_email_template_name = "hidp/accounts/verification/email/verification_body.html"

    def __init__(self, user, *, base_url, post_verification_redirect=None):
        super().__init__(base_url=base_url)
        self.user = user
        self.post_verification_redirect = post_verification_redirect

    def get_verification_url(self):
        return urljoin(
            self.base_url,
            email_verification.get_verify_email_url(
                self.user,
                next_url=self.post_verification_redirect,
            ),
        )

    def get_context(self, extra_context=None):
        return super().get_context(
            {
                "user": self.user,
                "verification_url": self.get_verification_url(),
            }
            | (extra_context or {})
        )

    def get_recipients(self):
        return [self.user.email]


class AccountExistsMailer(BaseMailer):
    subject_template_name = (
        "hidp/accounts/verification/email/account_exists_subject.txt"
    )
    email_template_name = "hidp/accounts/verification/email/account_exists_body.txt"
    html_email_template_name = (
        "hidp/accounts/verification/email/account_exists_body.html"
    )

    def __init__(self, user, *, base_url):
        super().__init__(base_url=base_url)
        self.user = user

    def get_password_reset_url(self):
        return urljoin(self.base_url, reverse("hidp_accounts:password_reset_request"))

    def get_context(self, extra_context=None):
        return super().get_context(
            {
                "user": self.user,
                "password_reset_url": self.get_password_reset_url(),
            }
            | (extra_context or {})
        )

    def get_recipients(self):
        return [self.user.email]


class PasswordResetRequestMailer(BaseMailer):
    subject_template_name = "hidp/accounts/recovery/email/password_reset_subject.txt"
    email_template_name = "hidp/accounts/recovery/email/password_reset_body.txt"
    html_email_template_name = "hidp/accounts/recovery/email/password_reset_body.html"
    token_generator = default_token_generator

    def __init__(self, user, *, base_url):
        super().__init__(base_url=base_url)
        self.user = user

    def get_password_reset_url(self):
        """
        Return the URL to the password reset page for the given user.

        Returns:
            An absolute URL to the password reset page for the given user.
        """
        return urljoin(
            self.base_url,
            reverse(
                "hidp_accounts:password_reset",
                kwargs={
                    "uidb64": urlsafe_base64_encode(force_bytes(self.user.pk)),
                    "token": self.token_generator.make_token(self.user),
                },
            ),
        )

    def get_context(self, extra_context=None):
        email_field_name = self.user.__class__.get_email_field_name()
        user_email = getattr(self.user, email_field_name)
        return super().get_context(
            {
                "email": user_email,
                "user": self.user,
                "password_reset_url": self.get_password_reset_url(),
            }
            | (extra_context or {})
        )

    def get_recipients(self):
        return [self.user.email]


class SetPasswordMailer(BaseMailer):
    subject_template_name = "hidp/accounts/recovery/email/set_password_subject.txt"
    email_template_name = "hidp/accounts/recovery/email/set_password_body.txt"
    html_email_template_name = "hidp/accounts/recovery/email/set_password_body.html"

    def __init__(self, user, *, base_url):
        super().__init__(base_url=base_url)
        self.user = user

    def get_set_password_url(self):
        return urljoin(self.base_url, reverse("hidp_account_management:set_password"))

    def get_context(self, extra_context=None):
        return super().get_context(
            {
                "user": self.user,
                "set_password_url": self.get_set_password_url(),
            }
            | (extra_context or {})
        )

    def get_recipients(self):
        return [self.user.email]


class PasswordChangedMailer(BaseMailer):
    subject_template_name = (
        "hidp/accounts/management/email/password_changed_subject.txt"
    )
    email_template_name = "hidp/accounts/management/email/password_changed_body.txt"
    html_email_template_name = (
        "hidp/accounts/management/email/password_changed_body.html"
    )

    def __init__(self, user, *, base_url):
        super().__init__(base_url=base_url)
        self.user = user

    def get_password_reset_url(self):
        return urljoin(self.base_url, reverse("hidp_accounts:password_reset_request"))

    def get_context(self, extra_context=None):
        return super().get_context(
            {
                "user": self.user,
                "password_reset_url": self.get_password_reset_url(),
            }
            | (extra_context or {})
        )

    def get_recipients(self):
        return [self.user.email]


class EmailChangeRequestMailer(BaseMailer):
    subject_template_name = "hidp/accounts/management/email/email_change_subject.txt"
    email_template_name = "hidp/accounts/management/email/email_change_body.txt"
    html_email_template_name = "hidp/accounts/management/email/email_change_body.html"

    def __init__(self, user, *, base_url, email_change_request, recipient):
        super().__init__(base_url=base_url)
        self.user = user
        self.email_change_request = email_change_request
        self.recipient = recipient

    def get_confirmation_url(self):
        return urljoin(
            self.base_url,
            reverse(
                "hidp_account_management:email_change_confirm",
                kwargs={
                    "token": (
                        tokens.email_change_token_generator.make_token(
                            str(self.email_change_request.pk), self.recipient
                        )
                    )
                },
            ),
        )

    def get_cancel_url(self):
        return urljoin(
            self.base_url,
            reverse("hidp_account_management:email_change_cancel"),
        )

    def get_context(self, extra_context=None):
        return super().get_context(
            {
                "user": self.user,
                "recipient": self.recipient,
                "current_email": self.email_change_request.current_email,
                "proposed_email": self.email_change_request.proposed_email,
                "confirmation_url": self.get_confirmation_url(),
                "cancel_url": self.get_cancel_url(),
            }
            | (extra_context or {})
        )

    def get_recipients(self):
        match self.recipient:
            case Recipient.CURRENT_EMAIL:
                return [self.email_change_request.current_email]
            case Recipient.PROPOSED_EMAIL:
                return [self.email_change_request.proposed_email]
            case _:
                raise ValueError(f"Invalid recipient: {self.recipient!r}")


class ProposedEmailExistsMailer(EmailChangeRequestMailer):
    subject_template_name = (
        "hidp/accounts/management/email/proposed_email_exists_subject.txt"
    )
    email_template_name = (
        "hidp/accounts/management/email/proposed_email_exists_body.txt"
    )
    html_email_template_name = (
        "hidp/accounts/management/email/proposed_email_exists_body.html"
    )

    def get_context(self, extra_context=None):
        return {
            "current_email": self.email_change_request.current_email,
            "proposed_email": self.email_change_request.proposed_email,
            "cancel_url": self.get_cancel_url(),
        }


class EmailChangedMailer(BaseMailer):
    subject_template_name = "hidp/accounts/management/email/email_changed_subject.txt"
    email_template_name = "hidp/accounts/management/email/email_changed_body.txt"
    html_email_template_name = "hidp/accounts/management/email/email_changed_body.html"

    def __init__(self, user, *, email_change_request, base_url):
        super().__init__(base_url=base_url)
        self.user = user
        self.email_change_request = email_change_request

    def get_context(self, extra_context=None):
        return super().get_context(
            {
                "current_email": self.email_change_request.current_email,
                "proposed_email": self.email_change_request.proposed_email,
            }
            | (extra_context or {})
        )

    def get_recipients(self):
        return [
            self.email_change_request.current_email,
            self.email_change_request.proposed_email,
        ]



from django.urls import path

from . import views

app_name = "hidp_otp_management"

urlpatterns = [
    path(
        "",
        views.OTPOverviewView.as_view(),
        name="manage",
    ),
    path(
        "disable/",
        views.OTPDisableView.as_view(),
        name="disable",
    ),
    path(
        "disable/recovery-code/",
        views.OTPDisableRecoveryCodesView.as_view(),
        name="disable-recovery-code",
    ),
    path(
        "recovery-codes/",
        views.OTPRecoveryCodesView.as_view(),
        name="recovery-codes",
    ),
    path(
        "setup/done/",
        views.OTPSetupDeviceDoneView.as_view(),
        name="setup-device-done",
    ),
    path(
        "setup/",
        views.OTPSetupDeviceView.as_view(),
        name="setup",
    ),
]



import random
import time

from unittest import mock

import requests

from jwcrypto import jwt

from django.test import RequestFactory, SimpleTestCase, TestCase

from hidp import config as hidp_config
from hidp.federated.constants import OIDC_STATES_SESSION_KEY
from hidp.federated.oidc import authorization_code_flow, exceptions

from ...test_federated.test_providers.example import (
    ExampleOIDCClient,
    code_challenge_from_code_verifier,
)


class NoPKCEOIDCClient(ExampleOIDCClient):
    has_pkce_support = False


class TestAuthenticationRequestParams(SimpleTestCase):
    def test_defaults(self):
        """Passing only the required parameters results in sensible defaults."""
        params = authorization_code_flow.get_authentication_request_parameters(
            client_id="client_id",
            redirect_uri="redirect_uri",
            state="state",
        )
        self.assertEqual(
            {
                "response_type": "code",
                "client_id": "client_id",
                "scope": "openid email profile",
                "state": "state",
                "redirect_uri": "redirect_uri",
            },
            params,
        )

    def test_custom_scope(self):
        """Custom scopes are passed through."""
        params = authorization_code_flow.get_authentication_request_parameters(
            client_id="client_id",
            redirect_uri="redirect_uri",
            state="state",
            scope="openid email",
        )
        self.assertEqual(
            {
                "client_id": "client_id",
                "redirect_uri": "redirect_uri",
                "response_type": "code",
                "scope": "openid email",
                "state": "state",
            },
            params,
        )

    def test_additional_parameters(self):
        """Additional parameters are passed through."""
        params = authorization_code_flow.get_authentication_request_parameters(
            client_id="client_id",
            redirect_uri="redirect_uri",
            state="state",
            ui_locales="nl",
        )
        self.assertEqual(
            {
                "response_type": "code",
                "client_id": "client_id",
                "scope": "openid email profile",
                "redirect_uri": "redirect_uri",
                "state": "state",
                "ui_locales": "nl",
            },
            params,
        )

    def test_override_response_type(self):
        """Overriding the default response type is not allowed."""
        # The only valid value for the code flow is "code".
        params = authorization_code_flow.get_authentication_request_parameters(
            client_id="client_id",
            redirect_uri="redirect_uri",
            state="state",
            response_type="token",
        )
        self.assertEqual(
            {
                "response_type": "code",
                "client_id": "client_id",
                "scope": "openid email profile",
                "redirect_uri": "redirect_uri",
                "state": "state",
            },
            params,
        )


class TestPrepareAuthenticationRequest(TestCase):
    def setUp(self):
        self.request = RequestFactory().get("/auth/")
        self.request.session = self.client.session

    def test_no_pkce_support(self):
        """Omits PKCE parameters when the client doesn't support it."""
        client = NoPKCEOIDCClient(client_id="client_id", client_secret="client_secret")
        url = authorization_code_flow.prepare_authentication_request(
            self.request, client=client, callback_url="/redirect/"
        )
        # Adds state to session
        self.assertIn(OIDC_STATES_SESSION_KEY, self.request.session)
        state_key = next(iter(self.request.session[OIDC_STATES_SESSION_KEY].keys()))
        # Adds correct parameters to URL
        self.assertEqual(
            f"https://example.com/auth"
            f"?response_type=code"
            f"&client_id={client.client_id}"
            f"&scope=openid+email+profile"
            f"&redirect_uri=http%3A%2F%2Ftestserver%2Fredirect%2F"
            f"&state={state_key}",
            url,
        )

    def test_prepare_no_callback_base_url(self):
        """Uses the client's authorization endpoint and the request's domain."""
        client = ExampleOIDCClient(client_id="client_id")
        url = authorization_code_flow.prepare_authentication_request(
            self.request, client=client, callback_url="/redirect/"
        )
        # Adds state to session
        self.assertIn(OIDC_STATES_SESSION_KEY, self.request.session)
        state_key = next(iter(self.request.session[OIDC_STATES_SESSION_KEY].keys()))
        # Adds code_verifier to session
        self.assertIn(
            "code_verifier", self.request.session[OIDC_STATES_SESSION_KEY][state_key]
        )
        code_verifier = self.request.session[OIDC_STATES_SESSION_KEY][state_key][
            "code_verifier"
        ]
        code_challenge = code_challenge_from_code_verifier(code_verifier)
        # Adds correct parameters to URL
        self.assertEqual(
            (
                f"https://example.com/auth"
                f"?response_type=code"
                f"&client_id={client.client_id}"
                f"&scope=openid+email+profile"
                f"&redirect_uri=http%3A%2F%2Ftestserver%2Fredirect%2F"
                f"&state={state_key}"
                f"&code_challenge={code_challenge}"
                f"&code_challenge_method=S256"
            ),
            url,
        )

    def test_prepare_callback_base_url(self):
        """Uses the client's authorization endpoint and callback base URL."""
        client = ExampleOIDCClient(
            client_id="client_id",
            callback_base_url="https://example.com/",
        )
        url = authorization_code_flow.prepare_authentication_request(
            self.request, client=client, callback_url="/redirect/"
        )
        # Adds state to session
        self.assertIn(OIDC_STATES_SESSION_KEY, self.request.session)
        state_key = next(iter(self.request.session[OIDC_STATES_SESSION_KEY]))
        # Adds code_verifier to session
        self.assertIn(
            "code_verifier", self.request.session[OIDC_STATES_SESSION_KEY][state_key]
        )
        code_verifier = self.request.session[OIDC_STATES_SESSION_KEY][state_key][
            "code_verifier"
        ]
        code_challenge = code_challenge_from_code_verifier(code_verifier)
        # Adds correct parameters to URL
        self.assertEqual(
            (
                f"https://example.com/auth"
                f"?response_type=code"
                f"&client_id={client.client_id}"
                f"&scope=openid+email+profile"
                f"&redirect_uri=https%3A%2F%2Fexample.com%2Fredirect%2F"
                f"&state={state_key}"
                f"&code_challenge={code_challenge}"
                f"&code_challenge_method=S256"
            ),
            url,
        )

    def test_create_pkce_challenge_no_state(self):
        """It is not possible to create a PKCE challenge before adding a state."""
        with self.assertRaisesMessage(
            ValueError,
            "Missing state in session. State must be added before"
            " creating a PKCE challenge.",
        ):
            authorization_code_flow.create_pkce_challenge(
                self.request, state_key="fake_state"
            )

    def test_limit_concurrent_authentication_requests(self):
        """The number of concurrent authentication requests (states) is limited."""
        # States *without* creation time (precaution for old/invalid states)
        states = [{} for _ in range(25)]
        # States with a creation time in the past
        states += [{"created_at": time.time() - n * 100} for n in range(25)]
        # Shuffle the states so they are in random order
        random.shuffle(states)
        # Add the states to the session
        self.request.session[OIDC_STATES_SESSION_KEY] = {
            f"state-{n}": state for n, state in enumerate(states)
        }
        authorization_code_flow.prepare_authentication_request(
            self.request,
            client=ExampleOIDCClient(client_id="client_id"),
            callback_url="/redirect/",
        )
        # Total amount of states is limited
        self.assertEqual(25, len(self.request.session[OIDC_STATES_SESSION_KEY]))
        # New state is the first one
        self.assertNotIn(
            next(iter(self.request.session[OIDC_STATES_SESSION_KEY].values())),
            states,
        )
        # Oldest states are removed (states without a creation time are considered old).
        # States are sorted by creation time, so the last 24 states are the newest.
        self.assertSequenceEqual(
            sorted(states, key=lambda state: state.get("created_at", 0), reverse=True)[
                :24
            ],
            list(self.request.session[OIDC_STATES_SESSION_KEY].values())[1:],
        )


class TestValidateAuthenticationCallback(TestCase):
    def setUp(self):
        self.session = self.client.session
        self.states = {"state-123": {"test": "test", "created_at": time.time()}}
        self.session[OIDC_STATES_SESSION_KEY] = self.states.copy()

    def test_missing_params(self):
        """Raises an OIDCError when the code and state are missing."""
        request = RequestFactory().get("/callback/")
        request.session = self.session
        with self.assertRaisesMessage(
            exceptions.OIDCError,
            "Missing 'code' in the authentication response",
        ):
            authorization_code_flow.validate_authentication_callback(request)
        # There's no way to know which state was used, so it's **not** removed.
        self.assertIn("state-123", request.session[OIDC_STATES_SESSION_KEY])

    def test_missing_code_with_state(self):
        """Raises an OIDCError when the code is missing."""
        request = RequestFactory().get("/callback/?state=state-123")
        request.session = self.session
        with self.assertRaisesMessage(
            exceptions.OIDCError,
            "Missing 'code' in the authentication response",
        ):
            authorization_code_flow.validate_authentication_callback(request)
        # The state is removed, as the authentication failed.
        self.assertNotIn("state-123", request.session[OIDC_STATES_SESSION_KEY])

    def test_missing_state_with_code(self):
        """Raises an OIDCError when the state is missing."""
        request = RequestFactory().get("/callback/?code=code&state=")
        request.session = self.session
        with self.assertRaisesMessage(
            exceptions.OIDCError,
            "Missing 'state' in the authentication response",
        ):
            authorization_code_flow.validate_authentication_callback(request)
        # There's no way to know which state was used, so it's **not** removed.
        self.assertIn("state-123", request.session[OIDC_STATES_SESSION_KEY])

    def test_invalid_state(self):
        """Raises an OIDCError when the state is invalid."""
        request = RequestFactory().get("/callback/?code=code&state=state-321")
        request.session = self.session
        with self.assertRaisesMessage(
            exceptions.OIDCError, "Invalid 'state' parameter"
        ):
            authorization_code_flow.validate_authentication_callback(request)
        # The state is **not** removed, as it's not the state from the request.
        self.assertIn("state-123", request.session[OIDC_STATES_SESSION_KEY])

    def test_expired_state(self):
        """Raises an OIDCError when the state has expired."""
        for state in [
            {},  # Missing created_at (precaution for old/invalid states)
            {"created_at": time.time() - 3600},  # Expired
        ]:
            with self.subTest(state=state):
                request = RequestFactory().get("/callback/?code=code&state=state-321")
                request.session = self.session
                request.session[OIDC_STATES_SESSION_KEY]["state-321"] = state
                with self.assertRaisesMessage(
                    exceptions.OIDCError, "Invalid 'state' parameter"
                ):
                    authorization_code_flow.validate_authentication_callback(request)
                # The state is removed, as the authentication failed.
                self.assertNotIn("state-321", request.session[OIDC_STATES_SESSION_KEY])
                # The other state is **not** removed.
                self.assertIn("state-123", request.session[OIDC_STATES_SESSION_KEY])

    def test_error_response(self):
        """Raises an OAuth2Error when the callback contains an error response."""
        request = RequestFactory().get(
            "/callback/?error=error&error_uri=https://example.com"
        )
        request.session = self.session
        with self.assertRaisesMessage(
            exceptions.OAuth2Error, "error (https://example.com)"
        ):
            authorization_code_flow.validate_authentication_callback(request)
        # There's no way to know which state was used, so it's **not** removed.
        self.assertIn("state-123", request.session[OIDC_STATES_SESSION_KEY])

    def test_error_response_with_state(self):
        """Raises an OAuth2Error when the callback contains an error response."""
        request = RequestFactory().get(
            "/callback/?error=error&error_description=description&state=state-123"
        )
        request.session = self.session
        with self.assertRaisesMessage(exceptions.OAuth2Error, "error: description"):
            authorization_code_flow.validate_authentication_callback(request)
        # The state is removed, as the authentication failed.
        self.assertNotIn("state-123", request.session[OIDC_STATES_SESSION_KEY])

    def test_valid_callback(self):
        """Validates the callback and returns the code and state."""
        request = RequestFactory().get("/callback/?code=code&state=state-123")
        request.session = self.session
        # Extracts code and state
        code, state = authorization_code_flow.validate_authentication_callback(request)
        self.assertEqual("code", code)
        self.assertEqual(self.states["state-123"], state)
        # Removes state from session
        self.assertNotIn(
            "state-123",
            request.session[OIDC_STATES_SESSION_KEY],
        )


@mock.patch.object(
    authorization_code_flow.requests,
    "post",
    autospec=True,
)
class TestObtainTokens(SimpleTestCase):
    def setUp(self):
        self.mock_response = {
            "access_token": "access_token",
            "id_token": "id_token",
            "token_type": "token_type",
        }

    def test_obtain_tokens_no_pkce(self, mock_requests_post):
        """Omits code_verifier when PKCE is not supported."""
        request = RequestFactory().get("/callback/")
        client = NoPKCEOIDCClient(client_id="client_id", client_secret="client_secret")
        mock_requests_post.return_value.json.return_value = self.mock_response

        tokens = authorization_code_flow.obtain_tokens(
            request=request,
            state={},
            client=client,
            code="code",
            callback_url="/redirect/",
        )

        mock_requests_post.assert_called_once_with(
            client.token_endpoint,
            data={
                "grant_type": "authorization_code",
                "code": "code",
                "redirect_uri": "http://testserver/redirect/",
                "client_id": client.client_id,
            },
            headers={
                "Accept": "application/json",
                "Origin": "http://testserver",
                "Authorization": "Basic Y2xpZW50X2lkOmNsaWVudF9zZWNyZXQ=",
            },
            timeout=(5, 30),
        )
        self.assertEqual(
            self.mock_response,
            tokens,
        )

    def test_obtain_tokens_no_pkce_or_basic_auth(self, mock_requests_post):
        """Passes the client_id and client_secret as form data."""
        request = RequestFactory().get("/callback/")
        client = NoPKCEOIDCClient(client_id="client_id", client_secret="client_secret")
        client.has_basic_auth_support = False
        mock_requests_post.return_value.json.return_value = self.mock_response

        tokens = authorization_code_flow.obtain_tokens(
            request=request,
            state={},
            client=client,
            code="code",
            callback_url="/redirect/",
        )

        mock_requests_post.assert_called_once_with(
            client.token_endpoint,
            data={
                "grant_type": "authorization_code",
                "code": "code",
                "redirect_uri": "http://testserver/redirect/",
                "client_id": client.client_id,
                "client_secret": client.client_secret,
            },
            headers={
                "Accept": "application/json",
                "Origin": "http://testserver",
            },
            timeout=(5, 30),
        )
        self.assertEqual(
            self.mock_response,
            tokens,
        )

    def test_no_code_verifier_in_state(self, mock_requests_post):
        """Raises an OIDCError when the code_verifier is missing from the state."""
        request = RequestFactory().get("/callback/")
        client = ExampleOIDCClient(client_id="client_id")
        with self.assertRaisesMessage(
            exceptions.OIDCError,
            "Missing 'code_verifier' in state.",
        ):
            authorization_code_flow.obtain_tokens(
                request=request,
                state={},
                client=client,
                code="code",
                callback_url="/redirect/",
            )

    def test_obtain_tokens_no_secret(self, mock_requests_post):
        """Obtains tokens without a client secret or callback base URL."""
        request = RequestFactory().get("/callback/")
        client = ExampleOIDCClient(client_id="client_id")
        mock_requests_post.return_value.json.return_value = self.mock_response

        tokens = authorization_code_flow.obtain_tokens(
            request=request,
            state={"code_verifier": "test"},
            client=client,
            code="code",
            callback_url="/redirect/",
        )

        mock_requests_post.assert_called_once_with(
            client.token_endpoint,
            data={
                "grant_type": "authorization_code",
                "code": "code",
                "redirect_uri": "http://testserver/redirect/",
                "client_id": client.client_id,
                "code_verifier": "test",
            },
            headers={
                "Accept": "application/json",
                "Origin": "http://testserver",
            },
            timeout=(5, 30),
        )
        self.assertEqual(
            self.mock_response,
            tokens,
        )

    def test_obtain_tokens_with_secret_and_callback_base_url(self, mock_requests_post):
        """Obtains tokens with a client secret and callback base URL."""
        request = RequestFactory().get("/callback/")
        client = ExampleOIDCClient(
            client_id="client_id",
            client_secret="client_secret",
            callback_base_url="https://example.com/",
        )
        mock_requests_post.return_value.json.return_value = self.mock_response

        tokens = authorization_code_flow.obtain_tokens(
            request=request,
            state={"code_verifier": "test"},
            client=client,
            code="code",
            callback_url="/redirect/",
        )

        mock_requests_post.assert_called_once_with(
            client.token_endpoint,
            data={
                "grant_type": "authorization_code",
                "code": "code",
                "redirect_uri": "https://example.com/redirect/",
                "client_id": client.client_id,
                "code_verifier": "test",
            },
            headers={
                "Accept": "application/json",
                "Origin": "https://example.com",
                "Authorization": "Basic Y2xpZW50X2lkOmNsaWVudF9zZWNyZXQ=",
            },
            timeout=(5, 30),
        )
        self.assertEqual(
            self.mock_response,
            tokens,
        )


@mock.patch.object(
    authorization_code_flow.jwks,
    "get_oidc_client_jwks",
    autospec=True,
)
class TestParseIdToken(TestCase):
    @classmethod
    def setUpTestData(cls):
        # Generate a key and a key set
        key = jwt.JWK(generate="RSA")
        key["kid"] = key.thumbprint()
        key_set = jwt.JWKSet()
        key_set.add(key)

        # Generate an unregistered key
        unregistered_key = jwt.JWK(generate="RSA")
        unregistered_key["kid"] = unregistered_key.thumbprint()

        # Expose the keys and key set
        cls.key = key
        cls.key_set = key_set
        cls.unregistered_key = unregistered_key

    def setUp(self):
        self.client = ExampleOIDCClient(client_id="client_id")
        hidp_config.configure_oidc_clients(self.client)

    def _get_token(self, *, claims=None, key=None, key_id=None):
        key = key or self.key
        token = jwt.JWT(
            header={"alg": "RS256", "kid": key_id or key["kid"]}, claims=claims or {}
        )
        token.make_signed_token(key)
        return token.serialize()

    def test_unable_to_get_jwks(self, mock_get_oidc_client_jwks):
        """Raises an OIDCError when the signing keys cannot be retrieved."""
        mock_get_oidc_client_jwks.return_value = None
        with self.assertRaisesMessage(
            exceptions.OIDCError,
            "Unable to get signing keys for 'example'."
            " The ID Token cannot be validated.",
        ):
            authorization_code_flow.parse_id_token(
                "id_token",
                client=self.client,
            )

    def test_invalid_token(self, mock_get_oidc_client_jwks):
        """Raises an OIDCError when the token is invalid."""
        with self.assertRaisesMessage(
            exceptions.OIDCError,
            "ID Token from 'example' is invalid.",
        ):
            authorization_code_flow.parse_id_token(
                "invalid_token",
                client=self.client,
            )

    def test_unknown_key(self, mock_get_oidc_client_jwks):
        """Raises an OIDCError when the key is not in the key set."""
        mock_get_oidc_client_jwks.return_value = self.key_set
        token = self._get_token(key=self.unregistered_key)

        with self.assertRaisesMessage(
            exceptions.OIDCError,
            "ID Token from 'example' is signed with an unknown key.",
        ):
            authorization_code_flow.parse_id_token(token, client=self.client)

    def test_missing_claims(self, mock_get_oidc_client_jwks):
        """Raises an OIDCError when the claims are missing."""
        mock_get_oidc_client_jwks.return_value = self.key_set
        token = self._get_token()

        with self.assertRaisesMessage(
            exceptions.OIDCError,
            "ID Token from 'example' has invalid or missing claims:"
            " Claim sub is missing.",
        ):
            authorization_code_flow.parse_id_token(token, client=self.client)

    def test_wrong_audience(self, mock_get_oidc_client_jwks):
        """Raises an OIDCError when the audience is incorrect."""
        mock_get_oidc_client_jwks.return_value = self.key_set
        token = self._get_token(
            claims={
                "sub": "subject",
                "iss": "issuer",
                "aud": "wrong_client_id",
            },
        )

        with self.assertRaisesMessage(
            exceptions.OIDCError,
            "ID Token from 'example' has invalid or missing claims:"
            " Invalid 'aud' value.",
        ):
            authorization_code_flow.parse_id_token(token, client=self.client)

    def test_expired_token(self, mock_get_oidc_client_jwks):
        """Raises an OIDCError when the token is expired."""
        mock_get_oidc_client_jwks.return_value = self.key_set
        token = self._get_token(
            claims={
                "sub": "subject",
                "iss": "issuer",
                "aud": "client_id",
                "exp": 1234567890,
            },
        )

        with self.assertRaisesMessage(
            exceptions.OIDCError,
            "ID Token from 'example' has expired: Expired at 1234567890",
        ):
            authorization_code_flow.parse_id_token(token, client=self.client)

    def test_not_yet_valid_token(self, mock_get_oidc_client_jwks):
        """Raises an OIDCError when the token is not yet valid."""
        mock_get_oidc_client_jwks.return_value = self.key_set
        token = self._get_token(
            claims={
                "sub": "subject",
                "iss": "issuer",
                "aud": "client_id",
                "exp": time.time() + 3600,
                "iat": time.time(),
                "nbf": time.time() + 300,
            },
        )

        with self.assertRaisesMessage(
            exceptions.OIDCError,
            "ID Token from 'example' is not yet valid.",
        ):
            authorization_code_flow.parse_id_token(token, client=self.client)

    def test_wrong_issuer(self, mock_get_oidc_client_jwks):
        """Raises an OIDCError when the issuer is incorrect."""
        mock_get_oidc_client_jwks.return_value = self.key_set
        token = self._get_token(
            claims={
                "sub": "subject",
                "iss": "https://example.test",
                "aud": "client_id",
                "exp": time.time() + 3600,
                "iat": time.time(),
            },
        )

        with self.assertRaisesMessage(
            exceptions.OIDCError,
            "ID Token from 'example' is not issued by 'https://example.com',"
            " got 'https://example.test'.",
        ):
            authorization_code_flow.parse_id_token(token, client=self.client)

    def test_unexpected_nonce(self, mock_get_oidc_client_jwks):
        """Raises an OIDCError when the nonce is incorrect."""
        mock_get_oidc_client_jwks.return_value = self.key_set
        token = self._get_token(
            claims={
                "sub": "subject",
                "iss": self.client.issuer,
                "aud": self.client.client_id,
                "exp": time.time() + 3600,
                "iat": time.time(),
                "nonce": "unexpected_nonce",
            },
        )

        with self.assertRaisesMessage(
            exceptions.OIDCError,
            "ID Token from 'example' contains an unexpected 'nonce' claim.",
        ):
            authorization_code_flow.parse_id_token(token, client=self.client)

    def test_parse_id_token(self, mock_get_oidc_client_jwks):
        """Parses the ID token and returns the claims."""
        mock_get_oidc_client_jwks.return_value = self.key_set
        claims = {
            "sub": "subject",
            "iss": self.client.issuer,
            "aud": self.client.client_id,
            "exp": time.time() + 3600,
            "iat": time.time(),
        }
        token = self._get_token(claims=claims)
        parsed_claims = authorization_code_flow.parse_id_token(
            token,
            client=self.client,
        )
        self.assertEqual(claims, parsed_claims)


@mock.patch.object(
    authorization_code_flow.requests,
    "get",
    autospec=True,
)
class TestGetUserInfo(TestCase):
    @staticmethod
    def _mock_response(content, *, status_code=200):
        response = requests.Response()
        response._content = content  # noqa: SLF001 (protected attribute)
        response.status_code = status_code
        return response

    def test_request_error(self, mock_requests_get):
        """Raises an OIDCError when the request fails."""
        client = ExampleOIDCClient(client_id="client_id")
        mock_requests_get.side_effect = requests.RequestException("Request failed.")
        with self.assertRaisesMessage(
            exceptions.OIDCError,
            "Failed to fetch user information from 'example'"
            " from 'https://example.com/userinfo'.",
        ):
            authorization_code_flow.get_user_info(
                client=client,
                access_token="access_token",
                claims={"sub": "subject"},
            )

    def test_response_error(self, mock_requests_get):
        """Raises an OIDCError when the response is not OK."""
        client = ExampleOIDCClient(client_id="client_id")
        mock_requests_get.return_value = self._mock_response(
            b"Not Found", status_code=404
        )
        with self.assertRaisesMessage(
            exceptions.OIDCError,
            "Error after fetching user information from 'example'"
            " from 'https://example.com/userinfo': 404.",
        ):
            authorization_code_flow.get_user_info(
                client=client,
                access_token="access_token",
                claims={"sub": "subject"},
            )

    def test_invalid_response(self, mock_requests_get):
        """Raises an OIDCError when the response is not JSON."""
        client = ExampleOIDCClient(client_id="client_id")
        mock_requests_get.return_value = self._mock_response(
            b"Not JSON, there must be a mistake."
        )
        with self.assertRaisesMessage(
            exceptions.OIDCError,
            "Failed to parse user information from 'example'"
            " from 'https://example.com/userinfo'.",
        ):
            authorization_code_flow.get_user_info(
                client=client,
                access_token="access_token",
                claims={"sub": "subject"},
            )

    def test_wrong_sub(self, mock_requests_get):
        """Raises an OIDCError when the "sub" claim doesn't match the expected value."""
        client = ExampleOIDCClient(client_id="client_id")
        mock_requests_get.return_value.json.return_value = {
            "sub": "wrong_subject",
        }
        with self.assertRaisesMessage(
            exceptions.OIDCError,
            "User information from 'example' does not match the ID token 'sub' claim.",
        ):
            authorization_code_flow.get_user_info(
                client=client,
                access_token="access_token",
                claims={"sub": "subject"},
            )

    def test_get_user_info(self, mock_requests_get):
        """Retrieves the user info and returns the claims."""
        client = ExampleOIDCClient(client_id="client_id")
        mock_requests_get.return_value.json.return_value = {
            "sub": "subject",
        }
        user_info = authorization_code_flow.get_user_info(
            client=client,
            access_token="access_token",
            claims={"sub": "subject"},
        )
        mock_requests_get.assert_called_once_with(
            client.userinfo_endpoint,
            headers={
                "Authorization": "Bearer access_token",
                "Accept": "application/json",
            },
            timeout=(5, 30),
        )
        self.assertEqual(
            user_info,
            {
                "sub": "subject",
            },
        )


class TestHandleAuthenticationCallback(TestCase):
    def setUp(self):
        self.request = RequestFactory().get("/callback/")
        self.request.session = self.client.session

    @mock.patch(
        "hidp.federated.oidc.authorization_code_flow.validate_authentication_callback",
        autospec=True,
        return_value=("code", {"code_verifier": "test"}),
    )
    @mock.patch(
        "hidp.federated.oidc.authorization_code_flow.obtain_tokens",
        autospec=True,
        return_value={
            "access_token": "access_token",
            "id_token": "id_token",
            "token_type": "token_type",
        },
    )
    @mock.patch(
        "hidp.federated.oidc.authorization_code_flow.parse_id_token",
        autospec=True,
        return_value={"claims": "claims"},
    )
    @mock.patch(
        "hidp.federated.oidc.authorization_code_flow.get_user_info",
        autospec=True,
        return_value={"user_info": "user_info"},
    )
    def test_handle_callback(
        self,
        mock_get_user_info,
        mock_parse_id_token,
        mock_obtain_tokens,
        mock_validate_callback,
    ):
        """Handles the authentication callback and returns the tokens."""
        client = ExampleOIDCClient(client_id="client_id")

        tokens, claims, user_info, _next_url = (
            authorization_code_flow.handle_authentication_callback(
                self.request,
                client=client,
                callback_url="/redirect/",
            )
        )

        mock_validate_callback.assert_called_once_with(self.request)
        mock_obtain_tokens.assert_called_once_with(
            self.request,
            state={"code_verifier": "test"},
            client=client,
            code="code",
            callback_url="/redirect/",
        )
        mock_parse_id_token.assert_called_once_with("id_token", client=client)
        mock_get_user_info.assert_called_once_with(
            client=client,
            access_token="access_token",
            claims={"claims": "claims"},
        )

        self.assertEqual(
            {
                "access_token": "access_token",
                "id_token": "id_token",
                "token_type": "token_type",
            },
            tokens,
        )

        self.assertEqual(
            {"claims": "claims"},
            claims,
        )

        self.assertEqual(
            {"user_info": "user_info"},
            user_info,
        )



from django.test import TestCase

from hidp.otp.exceptions import MultipleOtpDevicesError, NoOtpDeviceError
from hidp.otp.forms import VerifyTOTPForm
from hidp.test.factories import otp_factories, user_factories


class OTPVerifyFormTest(TestCase):
    def test_clean_no_device(self):
        user = user_factories.UserFactory()
        form = VerifyTOTPForm(user)
        form.cleaned_data = {
            "otp_token": "123456",
        }
        with self.assertRaises(NoOtpDeviceError):
            form.clean()

    def test_clean_with_device(self):
        user = user_factories.UserFactory()
        otp_factories.TOTPDeviceFactory(user=user, confirmed=True)
        form = VerifyTOTPForm(user)
        form.cleaned_data = {
            "otp_token": "123456",
        }

        # No exception should be raised
        self.assertFalse(form.is_valid(), "Expected form to be invalid")

    def test_clean_with_multiple_devices(self):
        user = user_factories.UserFactory()
        otp_factories.TOTPDeviceFactory.create_batch(2, user=user, confirmed=True)
        form = VerifyTOTPForm(user)
        form.cleaned_data = {
            "otp_token": "123456",
        }
        with self.assertRaises(MultipleOtpDevicesError):
            form.clean()



from enum import StrEnum, auto

OIDC_STATES_SESSION_KEY = "_hidp_oidc_states"


class OIDCError(StrEnum):
    ACCOUNT_EXISTS = auto()
    REQUEST_EXPIRED = auto()
    UNEXPECTED_ERROR = auto()
    INVALID_TOKEN = auto()
    INVALID_CREDENTIALS = auto()
    REGISTRATION_DISABLED = auto()



from django.test import TestCase, override_settings

from hidp.accounts import forms
from hidp.test.factories import user_factories


@override_settings(
    LANGUAGE_CODE="en",
    AUTHENTICATION_BACKENDS=["django.contrib.auth.backends.ModelBackend"],
)
class TestAuthenticationForm(TestCase):
    @classmethod
    def setUpTestData(cls):
        cls.user = user_factories.UserFactory()

    def test_is_valid(self):
        form = forms.AuthenticationForm(
            data={"username": self.user.email, "password": "P@ssw0rd!"}
        )
        self.assertTrue(form.is_valid())
        self.assertEqual(form.get_user(), self.user)

    def test_is_invalid(self):
        form = forms.AuthenticationForm(
            data={"username": self.user.email, "password": "wrong"}
        )
        self.assertFalse(form.is_valid())
        self.assertIsNone(form.get_user())
        self.assertFormError(
            form,
            None,
            [
                "Please enter a correct email address and password."
                " Note that both fields may be case-sensitive."
            ],
        )

    def test_is_inactive(self):
        self.user.is_active = False
        self.user.save()

        with self.subTest("Default backend"):
            form = forms.AuthenticationForm(
                data={"username": self.user.email, "password": "P@ssw0rd!"}
            )
            self.assertFalse(form.is_valid())
            # The user is not authenticated.
            self.assertIsNone(form.get_user())
            self.assertFormError(
                form,
                None,
                [
                    "Please enter a correct email address and password."
                    " Note that both fields may be case-sensitive."
                ],
            )

        with (
            self.subTest("AllowAllUsersModelBackend"),
            override_settings(
                AUTHENTICATION_BACKENDS=[
                    "django.contrib.auth.backends.AllowAllUsersModelBackend"
                ]
            ),
        ):
            form = forms.AuthenticationForm(
                data={"username": self.user.email, "password": "P@ssw0rd!"}
            )
            self.assertFalse(form.is_valid())
            # The user is authenticated, but not allowed to log in.
            self.assertEqual(form.get_user(), self.user)
            self.assertFormError(form, None, ["This account is inactive."])


class TestOptionalTOSUserCreationFormForm(TestCase):
    @classmethod
    def setUpTestData(cls):
        class NoTOSUserCreationForm(forms.UserCreationForm):
            """UserCreationForm without the agreed_to_tos field."""

            agreed_to_tos = None

        cls.form = NoTOSUserCreationForm

    def test_save(self):
        """Does not set agreed_to_tos on the user."""
        form = self.form(
            data={
                "email": "info@example.com",
                "password1": "P@ssw0rd!",
                "password2": "P@ssw0rd!",
            }
        )
        self.assertTrue(form.is_valid())
        user = form.save()
        self.assertIsNone(user.agreed_to_tos)



import logging

from datetime import timedelta

import requests

from jwcrypto import jwk

from django.core.cache import cache

from ...config import oidc_clients

_JWKS_CACHE_KEY_PREFIX = "hidp:oidc_client_jwks"
# Keys should rarely change. Keep this timeout high to avoid unnecessary requests
# to the JWKS endpoint during normal operation.
# Keys can be fetched eagerly (and periodically) if freshness is a concern.
_JWKS_DATA_CACHE_TIMEOUT = timedelta(days=7).total_seconds()
# If an error occurs while fetching the signing keys, cache the error to avoid
# repeated requests to the JWKS endpoint. The error may be temporary, so cache
# it for a short period of time.
_JWKS_ERROR_CACHE_TIMEOUT = timedelta(minutes=5).total_seconds()
# A sentinel value to indicate a cache miss.
_JWKS_DATA_MISSING = object()


logger = logging.getLogger(__name__)


def get_oidc_client_jwks(client, *, eager=False):
    """
    Retrieve the signing keys for an OIDC client by provider key.

    Arguments:
        client (OIDCClient):
            The provider key of the client to retrieve the signing keys for.
        eager (bool):
            Whether to bypass the cache and fetch the signing keys directly
            from the provider's JWKS endpoint.

            Use this to ensure that the keys are fresh and up-to-date.
            For example, on application startup or in a periodic task.

            Do not use this in the request/response cycle, it is a
            blocking operation that can significantly slow down the
            response time.

    Returns:
        jwcrypto.jwk.JWKSet | None:
            The signing keys for the OIDC client if available, or None if
            the keys could not be retrieved.

    Raises:
        KeyError:
            If the client is not registered.
    """
    # A client must be registered to retrieve its signing keys.
    # There is no reason to fetch keys for an unregistered client.
    if client != oidc_clients.get_oidc_client(client.provider_key):
        # Only trust the given client if it's the exact same instance.
        raise KeyError(f"Client is not registered for {client.provider_key!r}.")

    cache_key = f"{_JWKS_CACHE_KEY_PREFIX}:{client.provider_key}"

    # Check the cache, unless explicitly requested to bypass it.
    if not eager:
        jwks_data = cache.get(cache_key, default=_JWKS_DATA_MISSING)

        if jwks_data and jwks_data is not _JWKS_DATA_MISSING:
            try:
                # Return immediately if the cached data is valid.
                return jwk.JWKSet.from_json(jwks_data)
            except jwk.InvalidJWKValue:
                # The cached data could not be decoded. Very unlikely.
                # Fall through to fetch the signing keys from the provider's
                # JWKS endpoint.
                logger.exception(
                    "Failed to decode signing keys for %r from cache.",
                    client.provider_key,
                    extra={"jwks_data": jwks_data},
                )

        elif jwks_data is None:
            # The cache key is set to None, this indicates that a previous
            # attempt to fetch the signing keys failed. Return None to avoid
            # repeated requests to the JWKS endpoint.
            return None

        # The caller did not request eager fetching and the cache is empty.
        # Complain about the cache miss and reluctantly fetch the signing keys.
        # This is an unwanted situation and should be avoided.
        logger.warning(
            "Signing keys for %r are not cached, reluctantly fetching from %r.",
            client.provider_key,
            client.jwks_uri,
        )

    # Fetch the signing keys from the provider's JWKS endpoint.
    try:
        response = requests.get(
            client.jwks_uri,
            headers={
                "Accept": "application/json",
            },
            # Timeout in seconds
            timeout=(
                5,  # Connect timeout
                30,  # Read timeout
            ),
        )
    except requests.RequestException:
        # The request failed.
        logger.exception(
            "Failed to fetch signing keys for %r from %r.",
            client.provider_key,
            client.jwks_uri,
        )
        cache.set(cache_key, None, timeout=_JWKS_ERROR_CACHE_TIMEOUT)
        return None

    # Check the response status code.
    try:
        response.raise_for_status()
    except requests.RequestException:
        # The server returned an error response.
        logger.exception(
            "Error after fetching signing keys for %r from %r: %s.",
            client.provider_key,
            client.jwks_uri,
            response.status_code,
            extra={
                "response_reason": response.reason,
                "response_text": response.text,
            },
        )
        cache.set(cache_key, None, timeout=_JWKS_ERROR_CACHE_TIMEOUT)
        return None

    # The server returned a successful response.
    try:
        jwks = jwk.JWKSet.from_json(response.text)
    except jwk.JWException:
        # The response could not be decoded.
        logger.exception(
            "Failed to decode signing keys for %r from %r.",
            client.provider_key,
            client.jwks_uri,
            extra={
                "response_text": response.text,
            },
        )
        cache.set(cache_key, None, timeout=_JWKS_ERROR_CACHE_TIMEOUT)
        return None

    # Success! Cache the signing keys for future use.
    cache.set(cache_key, response.text, timeout=_JWKS_DATA_CACHE_TIMEOUT)
    return jwks


def refresh_registered_oidc_clients_jwks(stdout=None):
    """Refresh the signing keys for all registered OIDC clients."""
    for client in oidc_clients.get_registered_oidc_clients():
        if stdout:
            stdout.write(
                f"Fetching signing keys for {client.provider_key!r}"
                f" from {client.jwks_uri!r}..."
            )
        get_oidc_client_jwks(client, eager=True)



"""
OIDC Management URLs.

Provides the URL patterns for managing OIDC linked services.

Include this module in the root URL configuration:

    from hidp.federated import oidc_management_urls

    urlpatterns = [
        path("manage/oidc/", include(oidc_management_urls)),
    ]

This module also defines the namespace `hidp_oidc_management` for these URLs.

Include this namespace when reversing URLs, for example:

    reverse("hidp_oidc_management:linked_services")
"""

from django.urls import path

from . import views

app_name = "hidp_oidc_management"

urlpatterns = [
    path(
        "",
        views.OIDCLinkedServicesView.as_view(),
        name="linked_services",
    ),
    path(
        "link-account/",
        views.OIDCAccountLinkView.as_view(),
        name="link_account",
    ),
    path(
        "link-account/<str:provider_key>/done/",
        views.OIDCAccountLinkDoneView.as_view(),
        name="link_account_done",
    ),
    path(
        "unlink-account/<str:provider_key>/",
        views.OIDCAccountUnlinkView.as_view(),
        name="unlink_account",
    ),
    path(
        "unlink-account/<str:provider_key>/done/",
        views.OIDCAccountUnlinkDoneView.as_view(),
        name="unlink_account_done",
    ),
]



from django.apps import apps
from django.urls import include, path

from ..accounts import account_management_urls, account_urls
from ..federated import oidc_client_urls, oidc_management_urls

urlpatterns = [
    path("", include(account_urls)),
    path("login/oidc/", include(oidc_client_urls)),
    path("manage/", include(account_management_urls)),
    path("manage/oidc/", include(oidc_management_urls)),
]

if all(
    apps.is_installed(app)
    for app in (
        "hidp.otp",
        "django_otp.plugins.otp_totp",
        "django_otp.plugins.otp_static",
    )
):
    from ..otp import otp_management_urls

    urlpatterns += [
        path("manage/otp/", include(otp_management_urls)),
        path("otp/", include("hidp.otp.otp_urls")),
    ]

if apps.is_installed("hidp.oidc_provider"):
    urlpatterns += [
        path("o/", include("hidp.oidc_provider.urls")),
        path("api/", include("hidp.api.urls")),
    ]



"""Google OpenID Connect provider."""

from . import base

# Follow the instructions from Google on how to set up
# OpenID Connect on the Google Identity Platform:
# https://developers.google.com/identity/openid-connect/openid-connect
#
# Quick start:
#
# 1. Create a new project in the Google Cloud Console, and make sure to select it
#    as the active project. Skip this step if you already have a project.
#    https://console.cloud.google.com/projectcreate
#
# 2. Configure the OAuth consent screen. This is required for the OAuth 2.0 flow,
#    it is not possible to obtain OAuth 2.0 credentials without configuring it.
#    https://console.cloud.google.com/apis/credentials/consent
#
#    Note:
#    - Uploading a logo will require verification of the application.
#      Avoid uploading a logo if you are creating a test application.
#    - App domains and authorized domains are not required for testing.
#      They can be left empty.
#
# 3. Obtain OAuth 2.0 credentials:
#    https://console.cloud.google.com/apis/credentials/oauthclient
#
#    - Select "Web application" as the application type.
#    - Configure the authorized redirect URI:
#
#      https://<domain>/login/oidc/callback/google/
#
#      Note: The redirect URI must be a valid *public* URL or localhost.
#      For development use either localhost or a hosts file entry to
#      map a public domain to a local development server.
#
#      Tip: Use local.<production-domain> to avoid squatting on a real domain.
#   - Save the client ID and client secret, they will be required for the
#     OpenID Connect client configuration. Make sure to treat the client secret
#     as a secret, do not expose it in the source code or client-side code.


class GoogleOIDCClient(base.OIDCClient):
    provider_key = "google"

    # Google OpenID Connect configuration:
    # https://accounts.google.com/.well-known/openid-configuration
    issuer = "https://accounts.google.com"
    authorization_endpoint = "https://accounts.google.com/o/oauth2/v2/auth"
    token_endpoint = "https://oauth2.googleapis.com/token"  # noqa: S105 (not a secret)
    userinfo_endpoint = "https://openidconnect.googleapis.com/v1/userinfo"
    jwks_uri = "https://www.googleapis.com/oauth2/v3/certs"

    def __init__(self, *, client_id, client_secret, **kwargs):
        # Google requires both client_id and client_secret to be provided.
        super().__init__(client_id=client_id, client_secret=client_secret, **kwargs)



from django.test import TestCase, modify_settings

from hidp.accounts.auth import authenticate
from hidp.config import configure_oidc_clients
from hidp.federated.models import OpenIdConnection
from hidp.test.factories import user_factories

from ..test_providers.example import ExampleOIDCClient


@modify_settings(
    AUTHENTICATION_BACKENDS={"append": "hidp.federated.auth.backends.OIDCModelBackend"}
)
class TestOIDCModelBackend(TestCase):
    @classmethod
    def setUpTestData(cls):
        cls.user = user_factories.UserFactory()
        cls.oidc_connection = OpenIdConnection.objects.create(
            user=cls.user,
            provider_key="example",
            issuer_claim="https://example.com",
            subject_claim="example-subject",
        )
        # Assume the provider was in use at some point, but is now unregistered
        cls.unregistered_connection = OpenIdConnection.objects.create(
            user=cls.user,
            provider_key="unregistered",
            issuer_claim="https://example.com/other",
            subject_claim="other-subject",
        )

    def setUp(self):
        configure_oidc_clients(ExampleOIDCClient(client_id="example"))

    def test_missing_parameters(self):
        """Skip authentication if any of the required parameters is missing."""
        self.assertIsNone(
            authenticate(
                request=None, provider_key="example", subject_claim="example-subject"
            )
        )

    def test_unregistered_provider(self):
        """Return None if the provider is not registered (anymore), querying once."""
        with self.assertNumQueries(1):
            self.assertIsNone(
                authenticate(
                    request=None,
                    provider_key="unregistered",
                    issuer_claim="https://example.com/other",
                    subject_claim="other-subject",
                )
            )

    def test_invalid_credentials(self):
        """Return None if the credentials are invalid."""
        with self.assertNumQueries(1):
            self.assertIsNone(
                authenticate(
                    request=None,
                    provider_key="example",
                    issuer_claim="https://example.com",
                    subject_claim="invalid-subject",
                )
            )

    def test_success(self):
        """Return the user associated with the OpenID connection in one query."""
        with self.assertNumQueries(1):
            self.assertEqual(
                self.user,
                authenticate(
                    request=None,
                    provider_key="example",
                    issuer_claim="https://example.com",
                    subject_claim="example-subject",
                ),
            )



from rest_framework.routers import DefaultRouter

from django.urls import include, path

from .views import UserViewSet

router = DefaultRouter()
router.register("users", UserViewSet, basename="user")

app_name = "api"

urlpatterns = [
    path("", include(router.urls)),
]



from django_otp.plugins.otp_static.models import StaticDevice, StaticToken
from django_otp.plugins.otp_totp.models import TOTPDevice

from django.utils.translation import trans_null

from hidp.otp.exceptions import MultipleOtpDevicesError, NoOtpDeviceError

# The translation happens in the template, allowing the device names below to be
# translated to the user's language when they are displayed, but also allowing
# user-defined 'legacy' device names to be displayed as-is.
TOTP_DEVICE_NAME = trans_null.pgettext("OTP device name", "Authenticator app")
STATIC_DEVICE_NAME = trans_null.pgettext("OTP device name", "Recovery codes")

# Warning: changing the names above will result in the old strings from the
# database not being translated. If you need to change the names, you should
# add the old names to the list below, wrapped in
# trans_null.pgettext("OTP device name", ...).
_LEGACY_DEVICE_NAMES = [
    # trans_null.pgettext("OTP device name", "Old name"),  # noqa: ERA001
]


def get_or_create_devices(user):
    """
    Get or create OTP devices for a user.

    This function is used to ensure that a user has a TOTP device and a backup static
    device. If the user already has these devices, they are returned. If not, they are
    created in unconfirmed state.
    """
    totp_device, _created = TOTPDevice.objects.get_or_create(
        user=user,
        defaults={"name": TOTP_DEVICE_NAME, "confirmed": False},
    )
    static_device, backup_device_created = StaticDevice.objects.get_or_create(
        user=user,
        defaults={"name": STATIC_DEVICE_NAME, "confirmed": False},
    )
    if backup_device_created or not static_device.token_set.exists():
        reset_static_tokens(static_device)

    return totp_device, static_device


def reset_static_tokens(device, n=10):
    """
    Reset the static tokens for a device.

    This function deletes all existing static tokens for a device and creates 10 new
    ones. This amount should be sufficient for users to log in to disable MFA and
    during the time they have no access to their device but need to log in.
    """
    device.token_set.all().delete()
    for _ in range(n):
        device.token_set.create(token=StaticToken.random_token())


def get_device_for_user(user, device_class):
    """
    Get the confirmed device of a specific class for a user.

    It is expected that there is exactly one confirmed device of the specified class
    for the user. If there are none or more than one, an exception is raised.
    """
    try:
        return device_class.objects.devices_for_user(user, confirmed=True).get()
    except device_class.DoesNotExist as exc:
        raise NoOtpDeviceError from exc
    except device_class.MultipleObjectsReturned as exc:
        raise MultipleOtpDevicesError from exc



from django.urls import path

from hidp.otp import views

app_name = "hidp_otp"

urlpatterns = [
    path("verify/", views.VerifyTOTPView.as_view(), name="verify"),
    path(
        "verify/recovery-code/",
        views.VerifyRecoveryCodeView.as_view(),
        name="verify-recovery-code",
    ),
]



from django.core.management.base import BaseCommand

from ...oidc import jwks


class Command(BaseCommand):
    help = "Refresh the signing keys for all OIDC clients."

    def handle(self, *args, **options):
        jwks.refresh_registered_oidc_clients_jwks(stdout=self.stdout)



from functools import wraps


def otp_exempt(view_func):
    """Mark a view function as exempt from OTP verification."""

    @wraps(view_func)
    def wrapped_view(*args, **kwargs):
        return view_func(*args, **kwargs)

    wrapped_view.otp_exempt = True
    return wrapped_view



import uuid

from django.core import mail
from django.db import IntegrityError
from django.test import TestCase, override_settings

from hidp.test.factories import user_factories


@override_settings(DEFAULT_FROM_EMAIL="test@example.com")
class TestUserModel(TestCase):
    @classmethod
    def setUpTestData(cls):
        cls.user = user_factories.UserFactory(email="User@EXAMPLE.COM")

    def test_pk(self):
        """The primary key is a UUID version 7."""
        self.assertIsInstance(self.user.pk, uuid.UUID)
        self.assertEqual(7, self.user.pk.version)

    def test_unique_email(self):
        """The email field is unique, and case-insensitive."""
        with self.assertRaisesMessage(
            IntegrityError, "duplicate key value violates unique constraint"
        ):
            user_factories.UserFactory(email="USER@Example.COM")

    def test_change_email_unique_constraint(self):
        """Changing the email field should maintain the unique constraint."""
        user = user_factories.UserFactory()  # new user
        with self.assertRaisesMessage(
            IntegrityError, "duplicate key value violates unique constraint"
        ):
            user.email = "UseR@example.com"  # existing user's email, but different case
            user.save(update_fields=["email"])

    def test_str(self):
        """The string representation of the user is the (normalized) email address."""
        self.assertEqual("User@example.com", str(self.user))

    def test_set_password(self):
        """Sets the user's password to the hashed value of the raw password."""
        self.user.set_password("L3tM3In!")
        self.assertTrue(self.user.has_usable_password())
        self.assertTrue(self.user.check_password("L3tM3In!"))

    def test_set_unusable_password(self):
        """Sets the user's password to an unusable value."""
        self.user.set_unusable_password()
        self.assertFalse(self.user.has_usable_password())

    def test_email_user(self):
        """Sends an email to the user."""
        self.user.email_user("Test subject", "Test message")

        self.assertEqual(1, len(mail.outbox))
        email = mail.outbox[0]

        self.assertEqual([self.user.email], email.to)
        self.assertEqual("Test subject", email.subject)
        self.assertEqual("Test message", email.body)
        self.assertEqual("test@example.com", email.from_email)

    def test_get_short_name(self):
        """Returns the first name of the user."""
        user = user_factories.UserFactory(first_name="First", last_name="Last")
        self.assertEqual("First", user.get_short_name())
